#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from __future__ import with_statement
import atexit, os, sys, errno, inspect, re, datetime, platform, base64, signal, functools, time
try:
    import cPickle
except ImportError:
    import pickle as cPickle
if os.name == 'posix' and sys.version_info[0] < 3:
    try:
        import subprocess32 as subprocess
    except ImportError:
        import subprocess
else:
    import subprocess
try:
    TimeoutExpired = subprocess.TimeoutExpired
except AttributeError:

    class TimeoutExpired(Exception):
        pass


from collections import deque, defaultdict
try:
    import _winreg as winreg
except ImportError:
    try:
        import winreg
    except ImportError:
        winreg = None
from waflib import Errors
try:
    from hashlib import md5
except ImportError:
    try:
        from hashlib import sha1 as md5
    except ImportError:
        pass
else:
    try:
        md5().digest()
    except ValueError:
        from hashlib import sha1 as md5
try:
    import threading
except ImportError:
    if not 'JOBS' in os.environ:
        os.environ['JOBS'] = '1'

    class threading(object):
        pass

    class Lock(object):
        def acquire(self):
            pass

        def release(self):
            pass

    threading.Lock = threading.Thread = Lock
SIG_NIL = 'SIG_NIL_SIG_NIL_'.encode()
O644 = 420
O755 = 493
rot_chr = ['\\', '|', '/', '-']
rot_idx = 0


class ordered_iter_dict(dict):
    def __init__(self, *k, **kw):
        self.lst = deque()
        dict.__init__(self, *k, **kw)

    def clear(self):
        dict.clear(self)
        self.lst = deque()

    def __setitem__(self, key, value):
        if key in dict.keys(self):
            self.lst.remove(key)
        dict.__setitem__(self, key, value)
        self.lst.append(key)

    def __delitem__(self, key):
        dict.__delitem__(self, key)
        try:
            self.lst.remove(key)
        except ValueError:
            pass

    def __iter__(self):
        return reversed(self.lst)

    def keys(self):
        return reversed(self.lst)


class lru_node(object):
    __slots__ = ('next', 'prev', 'key', 'val')

    def __init__(self):
        self.next = self
        self.prev = self
        self.key = None
        self.val = None


class lru_cache(object):
    __slots__ = ('maxlen', 'table', 'head')

    def __init__(self, maxlen=100):
        self.maxlen = maxlen
        self.table = {}
        self.head = lru_node()
        self.head.next = self.head
        self.head.prev = self.head

    def __getitem__(self, key):
        node = self.table[key]
        if node is self.head:
            return node.val
        node.prev.next = node.next
        node.next.prev = node.prev
        node.next = self.head.next
        node.prev = self.head
        self.head = node.next.prev = node.prev.next = node
        return node.val

    def __setitem__(self, key, val):
        if key in self.table:
            node = self.table[key]
            node.val = val
            self.__getitem__(key)
        else:
            if len(self.table) < self.maxlen:
                node = lru_node()
                node.prev = self.head
                node.next = self.head.next
                node.prev.next = node.next.prev = node
            else:
                node = self.head = self.head.next
                try:
                    del self.table[node.key]
                except KeyError:
                    pass
            node.key = key
            node.val = val
            self.table[key] = node


class lazy_generator(object):
    def __init__(self, fun, params):
        self.fun = fun
        self.params = params

    def __iter__(self):
        return self

    def __next__(self):
        try:
            it = self.it
        except AttributeError:
            it = self.it = self.fun(*self.params)
        return next(it)

    next = __next__


is_win32 = os.sep == '\\' or sys.platform == 'win32' or os.name == 'nt'


def readf(fname, m='r', encoding='latin-1'):
    if sys.hexversion > 0x3000000 and not 'b' in m:
        m += 'b'
        with open(fname, m) as f:
            txt = f.read()
        if encoding:
            txt = txt.decode(encoding)
        else:
            txt = txt.decode()
    else:
        with open(fname, m) as f:
            txt = f.read()
    return txt


def writef(fname, data, m='w', encoding='latin-1'):
    if sys.hexversion > 0x3000000 and not 'b' in m:
        data = data.encode(encoding)
        m += 'b'
    with open(fname, m) as f:
        f.write(data)


def h_file(fname):
    m = md5()
    with open(fname, 'rb') as f:
        while fname:
            fname = f.read(200000)
            m.update(fname)
    return m.digest()


def readf_win32(f, m='r', encoding='latin-1'):
    flags = os.O_NOINHERIT | os.O_RDONLY
    if 'b' in m:
        flags |= os.O_BINARY
    if '+' in m:
        flags |= os.O_RDWR
    try:
        fd = os.open(f, flags)
    except OSError:
        raise IOError('Cannot read from %r' % f)
    if sys.hexversion > 0x3000000 and not 'b' in m:
        m += 'b'
        with os.fdopen(fd, m) as f:
            txt = f.read()
        if encoding:
            txt = txt.decode(encoding)
        else:
            txt = txt.decode()
    else:
        with os.fdopen(fd, m) as f:
            txt = f.read()
    return txt


def writef_win32(f, data, m='w', encoding='latin-1'):
    if sys.hexversion > 0x3000000 and not 'b' in m:
        data = data.encode(encoding)
        m += 'b'
    flags = os.O_CREAT | os.O_TRUNC | os.O_WRONLY | os.O_NOINHERIT
    if 'b' in m:
        flags |= os.O_BINARY
    if '+' in m:
        flags |= os.O_RDWR
    try:
        fd = os.open(f, flags)
    except OSError:
        raise OSError('Cannot write to %r' % f)
    with os.fdopen(fd, m) as f:
        f.write(data)


def h_file_win32(fname):
    try:
        fd = os.open(fname, os.O_BINARY | os.O_RDONLY | os.O_NOINHERIT)
    except OSError:
        raise OSError('Cannot read from %r' % fname)
    m = md5()
    with os.fdopen(fd, 'rb') as f:
        while fname:
            fname = f.read(200000)
            m.update(fname)
    return m.digest()


readf_unix = readf
writef_unix = writef
h_file_unix = h_file
if hasattr(os, 'O_NOINHERIT') and sys.hexversion < 0x3040000:
    readf = readf_win32
    writef = writef_win32
    h_file = h_file_win32
try:
    x = ''.encode('hex')
except LookupError:
    import binascii

    def to_hex(s):
        ret = binascii.hexlify(s)
        if not isinstance(ret, str):
            ret = ret.decode('utf-8')
        return ret
else:

    def to_hex(s):
        return s.encode('hex')


to_hex.__doc__ = """
Return the hexadecimal representation of a string

:param s: string to convert
:type s: string
"""


def listdir_win32(s):
    if not s:
        try:
            import ctypes
        except ImportError:
            return [x + ':\\' for x in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ']
        else:
            dlen = 4
            maxdrives = 26
            buf = ctypes.create_string_buffer(maxdrives * dlen)
            ndrives = ctypes.windll.kernel32.GetLogicalDriveStringsA(maxdrives * dlen, ctypes.byref(buf))
            return [str(buf.raw[4 * i:4 * i + 2].decode('ascii')) for i in range(int(ndrives / dlen))]
    if len(s) == 2 and s[1] == ":":
        s += os.sep
    if not os.path.isdir(s):
        e = OSError('%s is not a directory' % s)
        e.errno = errno.ENOENT
        raise e
    return os.listdir(s)


listdir = os.listdir
if is_win32:
    listdir = listdir_win32


def num2ver(ver):
    if isinstance(ver, str):
        ver = tuple(ver.split('.'))
    if isinstance(ver, tuple):
        ret = 0
        for i in range(4):
            if i < len(ver):
                ret += 256**(3 - i) * int(ver[i])
        return ret
    return ver


def to_list(val):
    if isinstance(val, str):
        return val.split()
    else:
        return val


def console_encoding():
    try:
        import ctypes
    except ImportError:
        pass
    else:
        try:
            codepage = ctypes.windll.kernel32.GetConsoleCP()
        except AttributeError:
            pass
        else:
            if codepage:
                return 'cp%d' % codepage
    return sys.stdout.encoding or ('cp1252' if is_win32 else 'latin-1')


def split_path_unix(path):
    return path.split('/')


def split_path_cygwin(path):
    if path.startswith('//'):
        ret = path.split('/')[2:]
        ret[0] = '/' + ret[0]
        return ret
    return path.split('/')


re_sp = re.compile('[/\\\\]+')


def split_path_win32(path):
    if path.startswith('\\\\'):
        ret = re_sp.split(path)[1:]
        ret[0] = '\\\\' + ret[0]
        if ret[0] == '\\\\?':
            return ret[1:]
        return ret
    return re_sp.split(path)


msysroot = None


def split_path_msys(path):
    if path.startswith(('/', '\\')) and not path.startswith(('//', '\\\\')):
        global msysroot
        if not msysroot:
            msysroot = subprocess.check_output(['cygpath', '-w', '/']).decode(sys.stdout.encoding or 'latin-1')
            msysroot = msysroot.strip()
        path = os.path.normpath(msysroot + os.sep + path)
    return split_path_win32(path)


if sys.platform == 'cygwin':
    split_path = split_path_cygwin
elif is_win32:
    if os.environ.get('MSYSTEM') and sys.executable.startswith('/'):
        split_path = split_path_msys
    else:
        split_path = split_path_win32
else:
    split_path = split_path_unix
split_path.__doc__ = """
Splits a path by / or \\; do not confuse this function with with ``os.path.split``

:type  path: string
:param path: path to split
:return:     list of string
"""


def check_dir(path):
    if not os.path.isdir(path):
        try:
            os.makedirs(path)
        except OSError as e:
            if not os.path.isdir(path):
                raise Errors.WafError('Cannot create the folder %r' % path, ex=e)


def check_exe(name, env=None):
    if not name:
        raise ValueError('Cannot execute an empty string!')

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(name)
    if fpath and is_exe(name):
        return os.path.abspath(name)
    else:
        env = env or os.environ
        for path in env['PATH'].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, name)
            if is_exe(exe_file):
                return os.path.abspath(exe_file)
    return None


def def_attrs(cls, **kw):
    for k, v in kw.items():
        if not hasattr(cls, k):
            setattr(cls, k, v)


def quote_define_name(s):
    fu = re.sub('[^a-zA-Z0-9]', '_', s)
    fu = re.sub('_+', '_', fu)
    fu = fu.upper()
    return fu


re_sh = re.compile('\\s|\'|"')


def shell_escape(cmd):
    if isinstance(cmd, str):
        return cmd
    return ' '.join(repr(x) if re_sh.search(x) else x for x in cmd)


def h_list(lst):
    return md5(repr(lst).encode()).digest()


if sys.hexversion < 0x3000000:

    def h_list_python2(lst):
        return md5(repr(lst)).digest()

    h_list_python2.__doc__ = h_list.__doc__
    h_list = h_list_python2


def h_fun(fun):
    try:
        return fun.code
    except AttributeError:
        if isinstance(fun, functools.partial):
            code = list(fun.args)
            code.extend(sorted(fun.keywords.items()))
            code.append(h_fun(fun.func))
            fun.code = h_list(code)
            return fun.code
        try:
            h = inspect.getsource(fun)
        except EnvironmentError:
            h = 'nocode'
        try:
            fun.code = h
        except AttributeError:
            pass
        return h


def h_cmd(ins):
    if isinstance(ins, str):
        ret = ins
    elif isinstance(ins, list) or isinstance(ins, tuple):
        ret = str([h_cmd(x) for x in ins])
    else:
        ret = str(h_fun(ins))
    if sys.hexversion > 0x3000000:
        ret = ret.encode('latin-1', 'xmlcharrefreplace')
    return ret


reg_subst = re.compile(r"(\\\\)|(\$\$)|\$\{([^}]+)\}")


def subst_vars(expr, params):
    def repl_var(m):
        if m.group(1):
            return '\\'
        if m.group(2):
            return '$'
        try:
            return params.get_flat(m.group(3))
        except AttributeError:
            return params[m.group(3)]

    return reg_subst.sub(repl_var, expr)


def destos_to_binfmt(key):
    if key == 'darwin':
        return 'mac-o'
    elif key in ('win32', 'cygwin', 'uwin', 'msys'):
        return 'pe'
    return 'elf'


def unversioned_sys_platform():
    s = sys.platform
    if s.startswith('java'):
        from java.lang import System
        s = System.getProperty('os.name')
        if s == 'Mac OS X':
            return 'darwin'
        elif s.startswith('Windows '):
            return 'win32'
        elif s == 'OS/2':
            return 'os2'
        elif s == 'HP-UX':
            return 'hp-ux'
        elif s in ('SunOS', 'Solaris'):
            return 'sunos'
        else:
            s = s.lower()
    if s == 'powerpc':
        return 'darwin'
    if s == 'win32' or s == 'os2':
        return s
    if s == 'cli' and os.name == 'nt':
        return 'win32'
    return re.split(r'\d+$', s)[0]


def nada(*k, **kw):
    pass


class Timer(object):
    def __init__(self):
        self.start_time = self.now()

    def __str__(self):
        delta = self.now() - self.start_time
        if not isinstance(delta, datetime.timedelta):
            delta = datetime.timedelta(seconds=delta)
        days = delta.days
        hours, rem = divmod(delta.seconds, 3600)
        minutes, seconds = divmod(rem, 60)
        seconds += delta.microseconds * 1e-6
        result = ''
        if days:
            result += '%dd' % days
        if days or hours:
            result += '%dh' % hours
        if days or hours or minutes:
            result += '%dm' % minutes
        return '%s%.3fs' % (result, seconds)

    def now(self):
        return datetime.datetime.utcnow()

    if hasattr(time, 'perf_counter'):

        def now(self):
            return time.perf_counter()


def read_la_file(path):
    sp = re.compile(r'^([^=]+)=\'(.*)\'$')
    dc = {}
    for line in readf(path).splitlines():
        try:
            _, left, right, _ = sp.split(line.strip())
            dc[left] = right
        except ValueError:
            pass
    return dc


def run_once(fun):
    cache = {}

    def wrap(*k):
        try:
            return cache[k]
        except KeyError:
            ret = fun(*k)
            cache[k] = ret
            return ret

    wrap.__cache__ = cache
    wrap.__name__ = fun.__name__
    return wrap


def get_registry_app_path(key, filename):
    if not winreg:
        return None
    try:
        result = winreg.QueryValue(key, "Software\\Microsoft\\Windows\\CurrentVersion\\App Paths\\%s.exe" % filename[0])
    except OSError:
        pass
    else:
        if os.path.isfile(result):
            return result


def lib64():
    if os.sep == '/':
        if platform.architecture()[0] == '64bit':
            if os.path.exists('/usr/lib64') and not os.path.exists('/usr/lib32'):
                return '64'
    return ''


def sane_path(p):
    return os.path.abspath(os.path.expanduser(p))


process_pool = []


def get_process():
    try:
        return process_pool.pop()
    except IndexError:
        filepath = os.path.dirname(os.path.abspath(__file__)) + os.sep + 'processor.py'
        cmd = [sys.executable, '-c', readf(filepath)]
        return subprocess.Popen(cmd, stdout=subprocess.PIPE, stdin=subprocess.PIPE, bufsize=0, close_fds=not is_win32)


def run_prefork_process(cmd, kwargs, cargs):
    if not kwargs.get('env'):
        kwargs['env'] = dict(os.environ)
    try:
        obj = base64.b64encode(cPickle.dumps([cmd, kwargs, cargs]))
    except (TypeError, AttributeError):
        return run_regular_process(cmd, kwargs, cargs)
    proc = get_process()
    if not proc:
        return run_regular_process(cmd, kwargs, cargs)
    proc.stdin.write(obj)
    proc.stdin.write('\n'.encode())
    proc.stdin.flush()
    obj = proc.stdout.readline()
    if not obj:
        raise OSError('Preforked sub-process %r died' % proc.pid)
    process_pool.append(proc)
    lst = cPickle.loads(base64.b64decode(obj))
    assert len(lst) == 5
    ret, out, err, ex, trace = lst
    if ex:
        if ex == 'OSError':
            raise OSError(trace)
        elif ex == 'ValueError':
            raise ValueError(trace)
        elif ex == 'TimeoutExpired':
            exc = TimeoutExpired(cmd, timeout=cargs['timeout'], output=out)
            exc.stderr = err
            raise exc
        else:
            raise Exception(trace)
    return ret, out, err


def lchown(path, user=-1, group=-1):
    if isinstance(user, str):
        import pwd
        entry = pwd.getpwnam(user)
        if not entry:
            raise OSError('Unknown user %r' % user)
        user = entry[2]
    if isinstance(group, str):
        import grp
        entry = grp.getgrnam(group)
        if not entry:
            raise OSError('Unknown group %r' % group)
        group = entry[2]
    return os.lchown(path, user, group)


def run_regular_process(cmd, kwargs, cargs={}):
    proc = subprocess.Popen(cmd, **kwargs)
    if kwargs.get('stdout') or kwargs.get('stderr'):
        try:
            out, err = proc.communicate(**cargs)
        except TimeoutExpired:
            if kwargs.get('start_new_session') and hasattr(os, 'killpg'):
                os.killpg(proc.pid, signal.SIGKILL)
            else:
                proc.kill()
            out, err = proc.communicate()
            exc = TimeoutExpired(proc.args, timeout=cargs['timeout'], output=out)
            exc.stderr = err
            raise exc
        status = proc.returncode
    else:
        out, err = (None, None)
        try:
            status = proc.wait(**cargs)
        except TimeoutExpired as e:
            if kwargs.get('start_new_session') and hasattr(os, 'killpg'):
                os.killpg(proc.pid, signal.SIGKILL)
            else:
                proc.kill()
            proc.wait()
            raise e
    return status, out, err


def run_process(cmd, kwargs, cargs={}):
    if kwargs.get('stdout') and kwargs.get('stderr'):
        return run_prefork_process(cmd, kwargs, cargs)
    else:
        return run_regular_process(cmd, kwargs, cargs)


def alloc_process_pool(n, force=False):
    global run_process, get_process, alloc_process_pool
    if not force:
        n = max(n - len(process_pool), 0)
    try:
        lst = [get_process() for x in range(n)]
    except OSError:
        run_process = run_regular_process
        get_process = alloc_process_pool = nada
    else:
        for x in lst:
            process_pool.append(x)


def atexit_pool():
    for k in process_pool:
        try:
            os.kill(k.pid, 9)
        except OSError:
            pass
        else:
            k.wait()


if (sys.hexversion < 0x207000f and not is_win32) or sys.hexversion >= 0x306000f:
    atexit.register(atexit_pool)
if os.environ.get('WAF_NO_PREFORK') or sys.platform == 'cli' or not sys.executable:
    run_process = run_regular_process
    get_process = alloc_process_pool = nada



#-----------------------------------------------------------------------------
# Copyright (c) 2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os
import sys
import argparse

SECRET = 'secret'

# Argument parser
parser = argparse.ArgumentParser(description="symbolic link collection test")
parser.add_argument(
    'test_type',
    type=str,
    help='Test type',
    choices=['samedir', 'subdir', 'parentdir'],
)
parser.add_argument(
    '--link-only',
    action='store_true',
    dest='link_only',
    default=False,
    help="Symlink-only test variant.",
)
args = parser.parse_args()

# Data dir
data_dir = os.path.join(sys._MEIPASS, 'data')

# Test modes / parametrization
if args.test_type == 'samedir':
    orig_file = None if args.link_only else os.path.join(data_dir, 'orig_file.txt')
    linked_file = os.path.join(data_dir, 'linked_file.txt')
elif args.test_type == 'subdir':
    orig_file = None if args.link_only else os.path.join(data_dir, 'subdir', 'orig_file.txt')
    linked_file = os.path.join(data_dir, 'linked_file.txt')
elif args.test_type == 'parentdir':
    orig_file = None if args.link_only else os.path.join(data_dir, 'orig_file.txt')
    linked_file = os.path.join(data_dir, 'subdir', 'linked_file.txt')

# Validation
if orig_file:
    # Check original file - must be a valid file and must not be a symbolic link
    print(f"Checking original file {orig_file}...", file=sys.stderr)
    assert os.path.isfile(orig_file), f"{orig_file} is not a file!"
    assert not os.path.islink(orig_file), f"{orig_file} is a symbolic link?!"
    with open(orig_file, 'r') as fp:
        assert fp.read() == SECRET, f"Invalid content in {orig_file}!"

    # Check linked file - must be a valid file (pointing to one) and must be a symbolic link
    print(f"Checking linked file {linked_file}...", file=sys.stderr)
    assert os.path.isfile(linked_file), f"{linked_file} is not a file!"
    assert os.path.islink(linked_file), f"{linked_file} is not a symlink!"
    with open(linked_file, 'r') as fp:
        assert fp.read() == SECRET, f"Invalid content in {linked_file}!"
else:
    # Check linked file - must be a valid file (pointing to one) and must NOT be a symbolic link (because original was
    # not collected)
    print(f"Checking only linked file {linked_file}...", file=sys.stderr)
    assert os.path.isfile(linked_file), f"{linked_file} is not a file!"
    assert not os.path.islink(linked_file), f"{linked_file} is a symbolic link, but it should not be!"
    with open(linked_file, 'r') as fp:
        assert fp.read() == SECRET, f"Invalid content in {linked_file}!"



#-----------------------------------------------------------------------------
# Copyright (c) 2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
#
# Basic tests for automatic binary vs. data file reclassification during anbalysis.

import pytest

import PyInstaller.utils.misc as miscutils
from PyInstaller.utils.tests import onedir_only


# Helpers
def _create_test_data_file(filename):
    filename.parent.mkdir(parents=True, exist_ok=True)
    # Create a text file
    filename.write_text("Test file", encoding='utf-8')


def _create_test_binary(filename):
    filename.parent.mkdir(parents=True, exist_ok=True)
    # Copy _ctypes extension
    import _ctypes
    import shutil
    shutil.copy2(_ctypes.__file__, filename)


def _create_test_build(pyi_builder, tmp_path, datas=None, binaries=None):
    extra_args = []
    for src_name, dest_name in datas or []:
        extra_args += ['--add-data', f"{src_name}:{dest_name}"]
    for src_name, dest_name in binaries or []:
        extra_args += ['--add-binary', f"{src_name}:{dest_name}"]

    pyi_builder.test_source("""
        print("Hello world!")
        """, pyi_args=extra_args)

    # Return path to the generated Analysis-XX.toc in the build directory
    analysis_toc_file = list((tmp_path / 'build' / 'test_source').glob("Analysis-*.toc"))
    assert len(analysis_toc_file) == 1
    analysis_toc_file = analysis_toc_file[0]

    # Load the serialized Analysis state, and take out the `binaries` and `datas` TOC lists.
    # The indices correspond to the lists' location in the `Analysis._guts`.
    analysis_data = miscutils.load_py_data_struct(analysis_toc_file)
    return (
        analysis_data[15],  # binaries
        analysis_data[18],  # datas
    )


# Test that we automatically reclassify a data file that was passed as a binary into its actual type.
@pytest.mark.linux
@pytest.mark.win32
@pytest.mark.darwin
@onedir_only
def test_automatic_reclassification_data_file(pyi_builder, tmp_path):
    binaries = []

    # Create test data file...
    src_path = tmp_path / 'test_file'
    _create_test_data_file(src_path)
    # ... and intentionally try to pass it as a binary.
    binaries.append((str(src_path), '.'))

    # Create test build and retrieve binaries and datas TOC lists
    binaries_toc, datas_toc = _create_test_build(pyi_builder, tmp_path, binaries=binaries)

    # We expect to find the test file's entry in the `datas` TOC list, and its typecode should be DATA.
    test_file_entries = [typecode for dest_name, src_name, typecode in binaries_toc if dest_name == 'test_file']
    assert len(test_file_entries) == 0

    test_file_entries = [typecode for dest_name, src_name, typecode in datas_toc if dest_name == 'test_file']
    assert len(test_file_entries) == 1
    assert test_file_entries[0] == 'DATA'


# Test that we automatically reclassify a binary that was passed as a data file into its actual type.
@pytest.mark.linux
@pytest.mark.win32
@pytest.mark.darwin
@onedir_only
def test_automatic_reclassification_binary(pyi_builder, tmp_path):
    datas = []

    # Create test binary...
    src_path = tmp_path / 'test_file'
    _create_test_binary(src_path)
    # ... and intentionally try to pass it as a data file.
    datas.append((str(src_path), '.'))

    # Create test build and retrieve binaries and datas TOC lists
    binaries_toc, datas_toc = _create_test_build(pyi_builder, tmp_path, datas=datas)

    # We expect to find the test file's entry in the `binaries` TOC list, and its typecode should be BINARY.
    test_file_entries = [typecode for dest_name, src_name, typecode in datas_toc if dest_name == 'test_file']
    assert len(test_file_entries) == 0

    test_file_entries = [typecode for dest_name, src_name, typecode in binaries_toc if dest_name == 'test_file']
    assert len(test_file_entries) == 1
    assert test_file_entries[0] == 'BINARY'



#-----------------------------------------------------------------------------
# Copyright (c) 2024, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import sys
import multiprocessing


def get_sys_meipass():
    return str(sys._MEIPASS)


def main(start_method):
    multiprocessing.set_start_method(start_method)

    # Start a pool with 4 worker processes.
    print('Running subprocess(es)...')
    with multiprocessing.Pool(processes=4) as pool:
        result = pool.apply_async(get_sys_meipass)
        subprocess_meipass = result.get()

        print(f"sys._MEIPASS in main process: {sys._MEIPASS}")
        print(f"sys._MEIPASS in sub-process: {subprocess_meipass}")

        assert sys._MEIPASS == subprocess_meipass


if __name__ == '__main__':
    multiprocessing.freeze_support()

    if len(sys.argv) != 2:
        print(f"Usage: {sys.argv[0]} <start-method>")
        sys.exit(1)

    main(sys.argv[1])



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib import Task


def build(bld):
    def run(self):
        for x in self.outputs:
            x.write('')

    for (name, cls) in Task.classes.items():
        cls.run = run



# Copyright (C) 2017-2023 PyInstaller Development Team.
# Copyright (C) 2016 Jason R Coombs <jaraco@jaraco.com>
#
# This file includes an almost complete copy of
# setuptools/extern/__init__.py, taken from setuptools 67.8.0.
# For PyInstaller the only change is to install a sub-class of VendorImporter,
# and that the source is auto-formatted using yapf.
#
# setuptools is licensed under the MIT License (expat) license:
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
# SPDX-License-Identifier: MIT

#--- Code of setuptools/extern/__init__.py starts here
import importlib.util
import sys


class VendorImporter:
    """
    A PEP 302 meta path importer for finding optionally-vendored or otherwise naturally-installed
    packages from root_name.
    """
    def __init__(self, root_name, vendored_names=(), vendor_pkg=None):
        self.root_name = root_name
        self.vendored_names = set(vendored_names)
        self.vendor_pkg = vendor_pkg or root_name.replace('extern', '_vendor')

    @property
    def search_path(self):
        """
        Search first the vendor package then as a natural package.
        """
        yield self.vendor_pkg + '.'
        yield ''

    def _module_matches_namespace(self, fullname):
        """Figure out if the target module is vendored."""
        root, base, target = fullname.partition(self.root_name + '.')
        return not root and any(map(target.startswith, self.vendored_names))

    def load_module(self, fullname):
        """
        Iterate over the search path to locate and load fullname.
        """
        root, base, target = fullname.partition(self.root_name + '.')
        for prefix in self.search_path:
            try:
                extant = prefix + target
                __import__(extant)
                mod = sys.modules[extant]
                sys.modules[fullname] = mod
                return mod
            except ImportError:
                pass
        else:
            raise ImportError(
                "The '{target}' package is required; normally this is bundled with this package so if you get "
                "this warning, consult the packager of your distribution.".format(**locals())
            )

    def create_module(self, spec):
        return self.load_module(spec.name)

    def exec_module(self, module):
        pass

    def find_spec(self, fullname, path=None, target=None):
        """Return a module spec for vendored names."""
        return importlib.util.spec_from_loader(fullname, self) if self._module_matches_namespace(fullname) else None

    def install(self):
        """
        Install this importer into sys.meta_path if not already present.
        """
        if self not in sys.meta_path:
            sys.meta_path.append(self)


# --- Code of setuptools/extern/__init__.py ends here


class MyVendorImporter(VendorImporter):
    @property
    def search_path(self):
        """
        Only search the vendor package, and not a natural package.
        """
        yield self.vendor_pkg + '.'


names = ('aaa', 'bbb', 'ccc')
MyVendorImporter(__name__, names).install()



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller.utils.cliutils import makespec


def test_maskespec_basic(tmp_path, monkeypatch):
    py_file = tmp_path / 'abcd.py'
    py_file.touch()

    monkeypatch.setattr('sys.argv', ['foobar', str(py_file)])
    # changing cwd does not work, since DEFAULT_SPECPATH is set *very* early
    monkeypatch.setattr('PyInstaller.building.makespec.DEFAULT_SPECPATH', str(tmp_path))
    makespec.run()

    spec_file = tmp_path / 'abcd.spec'
    assert spec_file.exists()
    spec_text = spec_file.read_text(encoding='utf-8')
    assert 'Analysis' in spec_text


def test_makespec_splash(tmp_path, monkeypatch):
    py_file = tmp_path / 'with_splash.py'
    py_file.touch()

    monkeypatch.setattr('sys.argv', ['foobar', '--splash', 'image.png', str(py_file)])
    monkeypatch.setattr('PyInstaller.building.makespec.DEFAULT_SPECPATH', str(tmp_path))
    makespec.run()

    spec_file = tmp_path / 'with_splash.spec'
    assert spec_file.exists()
    spec_text = spec_file.read_text('utf-8')
    assert 'Splash' in spec_text



#-----------------------------------------------------------------------------
# Copyright (c) 2013-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller.utils.hooks.tcl_tk import tcltk_info


def hook(hook_api):
    # Add all Tcl/Tk data files, based on the `TclTkInfo.data_files`. If Tcl/Tk is unavailable, the list is empty.
    #
    # NOTE: the list contains 3-element TOC tuples with full destination filenames (because other parts of code,
    # specifically splash-screen writer, currently require this format). Therefore, we need to use
    # `PostGraphAPI.add_datas` (which supports 3-element TOC tuples); if this was 2-element "hook-style" TOC list,
    #  we could just assign `datas` global hook variable, without implementing the post-graph `hook()` function.
    hook_api.add_datas(tcltk_info.data_files)



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import sys
from waflib.Tools import ccroot, ar, gxx
from waflib.Configure import conf


@conf
def find_icpc(conf):
    cxx = conf.find_program('icpc', var='CXX')
    conf.get_cc_version(cxx, icc=True)
    conf.env.CXX_NAME = 'icc'


def configure(conf):
    conf.find_icpc()
    conf.find_ar()
    conf.gxx_common_flags()
    conf.gxx_modifier_platform()
    conf.cxx_load_tools()
    conf.cxx_add_flags()
    conf.link_add_flags()



# ----------------------------------------------------------------------------
# Copyright (c) 2022-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# Qt modules information - the core of our Qt collection approach
# ----------------------------------------------------------------
#
# The python bindings for Qt (``PySide2``, ``PyQt5``, ``PySide6``, ``PyQt6``) consist of several python binary extension
# modules that provide bindings for corresponding Qt modules. For example, the ``PySide2.QtNetwork`` python extension
# module provides bindings for the ``QtNetwork`` Qt module from the ``qt/qtbase`` Qt repository.
#
# A Qt module can be considered as consisting of:
#  * a shared library (for example, on Linux, the shared library names for the ``QtNetwork`` Qt module in Qt5 and Qt6
#    are ``libQt5Network.so`` and ``libQt6Network.so``, respectively).
#  * plugins: a certain type (or class) of plugins is usually associated with a single Qt module (for example,
#    ``imageformats`` plugins are associated with the ``QtGui`` Qt module from the ``qt/qtbase`` Qt repository), but
#    additional plugins of that type may come from other Qt repositories. For example, ``imageformats/qsvg`` plugin
#    is provided by ``qtsvg/src/plugins/imageformats/svg`` from the ``qt/qtsvg`` repository, and ``imageformats/qpdf``
#    is provided by ``qtwebengine/src/pdf/plugins/imageformats/pdf`` from the ``qt/qtwebengine`` repository.
#  * translation files: names of translation files consist of a base name, which typically corresponds to the Qt
#    repository name, and language code. A single translation file usually covers all Qt modules contained within
#    the same repository. For example, translation files with base name ``qtbase``  contain translations for ``QtCore``,
#    ``QtGui``, ``QtWidgets``, ``QtNetwork``, and other Qt modules from the ``qt/qtbase`` Qt repository.
#
# The PyInstaller's built-in analysis of link-time dependencies ensures that when collecting a Qt python extension
# module, we automatically pick up the linked Qt shared libraries. However, collection of linked Qt shared libraries
# does not result in collection of plugins, nor translation files. In addition, the dependency of a Qt python extension
# module on other Qt python extension modules (i.e., at the bindings level) cannot be automatically determined due to
# PyInstaller's inability to scan imports in binary extensions.
#
# PyInstaller < 5.7 solved this problem using a dictionary that associated a Qt shared library name with python
# extension name, plugins, and translation files. For each hooked Qt python extension module, the hook calls a helper
# that analyzes the extension file for link-time dependencies, and matches those against the dictionary. Therefore,
# based on linked shared libraries, we could recursively infer the list of files to collect in addition to the shared
# libraries themselves:
#  - plugins and translation files belonging to Qt modules whose shared libraries we collect
#  - Qt python extension modules corresponding to the Qt modules that we collect
#
# The above approach ensures that even if analyzed python script contains only ``from PySide2 import QtWidgets``,
# we would also collect ``PySide2.QtGui`` and ``PySide2.QtCore``, as well as all corresponding Qt module files
# (the shared libraries, plugins, translation files). For this to work, a hook must be provided for the
# ``PySide2.QtWidgets`` that performs the recursive analysis of the extension module file; so to ensure that each
# Qt python extension module by itself ensures collection of all its dependencies, we need to hook all Qt python
# extension modules provided by specific python Qt bindings package.
#
# The above approach with single dictionary, however, has several limitations:
#  - it cannot provide association for Qt python module that binds a Qt module without a shared library (i.e., a
#    headers-only module, or a statically-built module). In such cases, potential plugins and translations should
#    be associated directly with the Qt python extension file instead of the Qt module's (non-existent) shared library.
#  - it cannot (directly) handle differences between Qt5 and Qt6; we had to build a second dictionary
#  - it cannot handle differences between the bindings themselves; for example, PyQt5 binds some Qt modules that
#    PySide2 does not bind. Or, the binding's Qt python extension module is named differently in PyQt and PySide
#    bindings (or just differently in PyQt5, while PySide2, PySide6, and PyQt6 use the same name).
#
# In order address the above shortcomings, we now store all information a list of structures that contain information
# for a particular Qt python extension and/or Qt module (shared library):
#  - python extension name (if applicable)
#  - Qt module name base (if applicable)
#  - plugins
#  - translation files base name
#  - applicable Qt version (if necessary)
#  - applicable Qt bindings (if necessary)
#
# This list is used to dynamically construct two dictionaries (based on the bindings name and Qt version):
#  - mapping python extension names to associated module information
#  - mapping Qt shared library names to associated module information
# This allows us to associate plugins and translations with either Qt python extension or with the Qt module's shared
# library (or both), whichever is applicable.
#
# The `qt_dynamic_dependencies_dict`_ from the original approach was constructed using several information sources, as
# documented `here
# <https://github.com/pyinstaller/pyinstaller/blob/fbf7948be85177dd44b41217e9f039e1d176de6b/PyInstaller/utils/hooks/qt.py#L266-L362>`_.
#
# In the current approach, the relations stored in the `QT_MODULES_INFO`_ list were determined directly, by inspecting
# the Qt source code. This requires some prior knowledge of how the Qt code is organized (repositories and individual Qt
# modules within them), as well as some searching based on guesswork. The procedure can be outlined as follows:
#  * check out the `main Qt repository <git://code.qt.io/qt/qt5.git>`_. This repository contains references to all other
#    Qt repositories in the form of git submodules.
#  * for Qt5:
#      * check out the latest release tag, e.g., v5.15.2, then check out the submodules.
#      * search the Qt modules' qmake .pro files; for example, ``qtbase/src/network/network.pro`` for QtNetwork module.
#        The plugin types associated with the module are listed in the ``MODULE_PLUGIN_TYPES`` variable (in this case,
#        ``bearer``).
#      * all translations are gathered in ``qttranslations`` sub-module/repository, and their association with
#        individual repositories can be seen in ``qttranslations/translations/translations.pro``.
#  * for Qt6:
#      * check out the latest release tag, e.g., v6.3.1, then check out the submodules.
#      * search the Qt modules' CMake files; for example, ``qtbase/src/network/CMakeLists.txt`` for QtNetwork module.
#        The plugin types associated with the module are listed under ``PLUGIN_TYPES`` argument of the
#        ``qt_internal_add_module()`` function that defines the Qt module.
#
# The idea is to make a list of all extension modules found in a Qt bindings package, as well as all available plugin
# directories (which correspond to plugin types) and translation files. For each extension, identify the corresponding
# Qt module (shared library name) and its associated plugins and translation files. Once this is done, most of available
# plugins and translations in the python bindings package should have a corresponding python Qt extension module
# available; this gives us associations based on the python extension module names as well as based on the Qt shared
# library names. For any plugins and translation files remaining unassociated, identify the corresponding Qt module;
# this gives us associations based only on Qt shared library names. While this second group of associations are never
# processed directly (due to lack of corresponding python extension), they may end up being processed during the
# recursive dependency analysis, if the corresponding Qt shared library is linked against by some Qt python extension
# or another Qt shared library.


# This structure is used to define Qt module information, such as python module/extension name, Qt module (shared
# library) name, translation files' base names, plugins, as well as associated python bindings (which implicitly
# also encode major Qt version).
class _QtModuleDef:
    def __init__(self, module, shared_lib=None, translations=None, plugins=None, bindings=None):
        # Python module (extension) name without package namespace. For example, `QtCore`.
        # Can be None if python bindings do not bind the module, but we still need to establish relationship between
        # the Qt module (shared library) and its plugins and translations.
        self.module = module
        # Associated Qt module (shared library), if any. Used during recursive dependency analysis, where a python
        # module (extension) is analyzed for linked Qt modules (shared libraries), and then their corresponding
        # python modules (extensions) are added to hidden imports. For example, the Qt module name is `Qt5Core` or
        # `Qt6Core`, depending on the Qt version. Can be None for python modules that are not tied to a particular
        # Qt shared library (for example, the corresponding Qt module is headers-only) and hence they cannot be
        # inferred from recursive link-time dependency analysis.
        self.shared_lib = shared_lib
        # List of base names of translation files (if any) associated with the Qt module. Multiple base names may be
        # associated with a single module.
        # For example, `['qt', 'qtbase']` for `QtCore` or `['qtmultimedia']` for `QtMultimedia`.
        self.translations = translations or []
        # List of plugins associated with the Qt module.
        self.plugins = plugins or []
        # List of bindings (PySide2, PyQt5, PySide6, PyQt6) that provide the python module. This allows association of
        # plugins and translations with shared libraries even for bindings that do not provide python module binding
        # for the Qt module.
        self.bindings = set(bindings or [])


# All Qt-based bindings.
ALL_QT_BINDINGS = {"PySide2", "PyQt5", "PySide6", "PyQt6"}

# Qt modules information - the core of our Qt collection approach.
#
# For every python module/extension (i.e., entry in the list below that has valid `module`), we need a corresponding
# hook, ensuring that the extension file is analyzed, so that we collect the associated plugins and translation
# files, as well as perform recursive analysis of link-time binary dependencies (so that plugins and translation files
# belonging to those dependencies are collected as well).
QT_MODULES_INFO = (
    # *** qt/qt3d ***
    _QtModuleDef("Qt3DAnimation", shared_lib="3DAnimation"),
    _QtModuleDef("Qt3DCore", shared_lib="3DCore"),
    _QtModuleDef("Qt3DExtras", shared_lib="3DExtras"),
    _QtModuleDef("Qt3DInput", shared_lib="3DInput", plugins=["3dinputdevices"]),
    _QtModuleDef("Qt3DLogic", shared_lib="3DLogic"),
    _QtModuleDef(
        "Qt3DRender", shared_lib="3DRender", plugins=["geometryloaders", "renderplugins", "renderers", "sceneparsers"]
    ),

    # *** qt/qtactiveqt ***
    # The python module is called QAxContainer in PyQt bindings, but QtAxContainer in PySide. The associated Qt module
    # is header-only, so there is no shared library.
    _QtModuleDef("QAxContainer", bindings=["PyQt*"]),
    _QtModuleDef("QtAxContainer", bindings=["PySide*"]),

    # *** qt/qtcharts ***
    # The python module is called QtChart in PyQt5, and QtCharts in PySide2, PySide6, and PyQt6 (which corresponds to
    # the associated Qt module name, QtCharts).
    _QtModuleDef("QtChart", shared_lib="Charts", bindings=["PyQt5"]),
    _QtModuleDef("QtCharts", shared_lib="Charts", bindings=["!PyQt5"]),

    # *** qt/qtbase ***
    # QtConcurrent python module is available only in PySide bindings.
    _QtModuleDef(None, shared_lib="Concurrent", bindings=["PyQt*"]),
    _QtModuleDef("QtConcurrent", shared_lib="Concurrent", bindings=["PySide*"]),
    _QtModuleDef("QtCore", shared_lib="Core", translations=["qt", "qtbase"]),
    # QtDBus python module is available in all bindings but PySide2.
    _QtModuleDef(None, shared_lib="DBus", bindings=["PySide2"]),
    _QtModuleDef("QtDBus", shared_lib="DBus", bindings=["!PySide2"]),
    # QtNetwork uses different plugins in Qt5 and Qt6.
    _QtModuleDef("QtNetwork", shared_lib="Network", plugins=["bearer"], bindings=["PySide2", "PyQt5"]),
    _QtModuleDef(
        "QtNetwork",
        shared_lib="Network",
        plugins=["networkaccess", "networkinformation", "tls"],
        bindings=["PySide6", "PyQt6"]
    ),
    _QtModuleDef(
        "QtGui",
        shared_lib="Gui",
        plugins=[
            "accessiblebridge",
            "egldeviceintegrations",
            "generic",
            "iconengines",
            "imageformats",
            "platforms",
            "platforms/darwin",
            "platforminputcontexts",
            "platformthemes",
            "xcbglintegrations",
            # The ``wayland-*`` plugins are part of QtWaylandClient Qt module, whose shared library
            # (e.g., libQt5WaylandClient.so) is linked by the wayland-related ``platforms`` plugins. Ideally, we would
            # collect these plugins based on the QtWaylandClient shared library entry, but as our Qt hook utilities do
            # not scan the plugins for dependencies, that would not work. So instead we list these plugins under QtGui
            # to achieve pretty much the same end result.
            "wayland-decoration-client",
            "wayland-graphics-integration-client",
            "wayland-shell-integration"
        ]
    ),
    _QtModuleDef("QtOpenGL", shared_lib="OpenGL"),
    # This python module is specific to PySide2 and has no associated Qt module.
    _QtModuleDef("QtOpenGLFunctions", bindings=["PySide2"]),
    # This Qt module was introduced with Qt6.
    _QtModuleDef("QtOpenGLWidgets", shared_lib="OpenGLWidgets", bindings=["PySide6", "PyQt6"]),
    _QtModuleDef("QtPrintSupport", shared_lib="PrintSupport", plugins=["printsupport"]),
    _QtModuleDef("QtSql", shared_lib="Sql", plugins=["sqldrivers"]),
    _QtModuleDef("QtTest", shared_lib="Test"),
    _QtModuleDef("QtWidgets", shared_lib="Widgets", plugins=["styles"]),
    _QtModuleDef("QtXml", shared_lib="Xml"),

    # *** qt/qtconnectivity ***
    _QtModuleDef("QtBluetooth", shared_lib="QtBluetooth", translations=["qtconnectivity"]),
    _QtModuleDef("QtNfc", shared_lib="Nfc", translations=["qtconnectivity"]),

    # *** qt/qtdatavis3d ***
    _QtModuleDef("QtDataVisualization", shared_lib="DataVisualization"),

    # *** qt/qtdeclarative ***
    _QtModuleDef("QtQml", shared_lib="Qml", translations=["qtdeclarative"], plugins=["qmltooling"]),
    # Have the Qt5 variant collect translations for qtquickcontrols (qt/qtquickcontrols provides only QtQuick plugins).
    _QtModuleDef(
        "QtQuick",
        shared_lib="Quick",
        translations=["qtquickcontrols"],
        plugins=["scenegraph"],
        bindings=["PySide2", "PyQt5"]
    ),
    _QtModuleDef("QtQuick", shared_lib="Quick", plugins=["scenegraph"], bindings=["PySide6", "PyQt6"]),
    # Qt6-only; in Qt5, this module is part of qt/qtquickcontrols2. Python module is available only in PySide6.
    _QtModuleDef(None, shared_lib="QuickControls2", bindings=["PyQt6"]),
    _QtModuleDef("QtQuickControls2", shared_lib="QuickControls2", bindings=["PySide6"]),
    _QtModuleDef("QtQuickWidgets", shared_lib="QuickWidgets"),

    # *** qt/qtgamepad ***
    # No python module; shared library -> plugins association entry.
    _QtModuleDef(None, shared_lib="Gamepad", plugins=["gamepads"]),

    # *** qt/qtgraphs ***
    # Qt6 >= 6.6.0; python module is available only in PySide6.
    _QtModuleDef("QtGraphs", shared_lib="Graphs", bindings=["PySide6"]),

    # *** qt/qthttpserver ***
    # Qt6 >= 6.4.0; python module is available only in PySide6.
    _QtModuleDef("QtHttpServer", shared_lib="HttpServer", bindings=["PySide6"]),

    # *** qt/qtlocation ***
    # QtLocation was reintroduced in Qt6 v6.5.0.
    _QtModuleDef(
        "QtLocation",
        shared_lib="Location",
        translations=["qtlocation"],
        plugins=["geoservices"],
        bindings=["PySide2", "PyQt5", "PySide6"]
    ),
    _QtModuleDef(
        "QtPositioning",
        shared_lib="Positioning",
        translations=["qtlocation"],
        plugins=["position"],
    ),

    # *** qt/qtmacextras ***
    # Qt5-only Qt module.
    _QtModuleDef("QtMacExtras", shared_lib="MacExtras", bindings=["PySide2", "PyQt5"]),

    # *** qt/qtmultimedia ***
    # QtMultimedia on Qt6 currently uses only a subset of plugin names from Qt5 counterpart.
    _QtModuleDef(
        "QtMultimedia",
        shared_lib="Multimedia",
        translations=["qtmultimedia"],
        plugins=[
            "mediaservice", "audio", "video/bufferpool", "video/gstvideorenderer", "video/videonode", "playlistformats",
            "resourcepolicy"
        ],
        bindings=["PySide2", "PyQt5"]
    ),
    _QtModuleDef(
        "QtMultimedia",
        shared_lib="Multimedia",
        translations=["qtmultimedia"],
        # `multimedia` plugins are available as of Qt6 >= 6.4.0; earlier versions had `video/gstvideorenderer` and
        # `video/videonode` plugins.
        plugins=["multimedia", "video/gstvideorenderer", "video/videonode"],
        bindings=["PySide6", "PyQt6"]
    ),
    _QtModuleDef("QtMultimediaWidgets", shared_lib="MultimediaWidgets"),
    # Qt6-only Qt module; python module is available in PySide6 >= 6.4.0 and PyQt6 >= 6.5.0
    _QtModuleDef("QtSpatialAudio", shared_lib="SpatialAudio", bindings=["PySide6", "PyQt6"]),

    # *** qt/qtnetworkauth ***
    # QtNetworkAuth python module is available in all bindings but PySide2.
    _QtModuleDef(None, shared_lib="NetworkAuth", bindings=["PySide2"]),
    _QtModuleDef("QtNetworkAuth", shared_lib="NetworkAuth", bindings=["!PySide2"]),

    # *** qt/qtpurchasing ***
    # Qt5-only Qt module, python module is available only in PyQt5.
    _QtModuleDef("QtPurchasing", shared_lib="Purchasing", bindings=["PyQt5"]),

    # *** qt/qtquick1 ***
    # This is an old, Qt 5.3-era module...
    _QtModuleDef(
        "QtDeclarative",
        shared_lib="Declarative",
        translations=["qtquick1"],
        plugins=["qml1tooling"],
        bindings=["PySide2", "PyQt5"]
    ),

    # *** qt/qtquick3d ***
    # QtQuick3D python module is available in all bindings but PySide2.
    _QtModuleDef(None, shared_lib="Quick3D", bindings=["PySide2"]),
    _QtModuleDef("QtQuick3D", shared_lib="Quick3D", bindings=["!PySide2"]),
    # No python module; shared library -> plugins association entry.
    _QtModuleDef(None, shared_lib="Quick3DAssetImport", plugins=["assetimporters"]),

    # *** qt/qtquickcontrols2 ***
    # Qt5-only module; in Qt6, this module is part of qt/declarative. Python module is available only in PySide2.
    _QtModuleDef(None, translations=["qtquickcontrols2"], shared_lib="QuickControls2", bindings=["PyQt5"]),
    _QtModuleDef(
        "QtQuickControls2", translations=["qtquickcontrols2"], shared_lib="QuickControls2", bindings=["PySide2"]
    ),

    # *** qt/qtremoteobjects ***
    _QtModuleDef("QtRemoteObjects", shared_lib="RemoteObjects"),

    # *** qt/qtscxml ***
    # Python module is available only in PySide bindings. Plugins are available only in Qt6.
    # PyQt wheels do not seem to ship the corresponding Qt modules (shared libs) at all.
    _QtModuleDef("QtScxml", shared_lib="Scxml", bindings=["PySide2"]),
    _QtModuleDef("QtScxml", shared_lib="Scxml", plugins=["scxmldatamodel"], bindings=["PySide6"]),
    # Qt6-only Qt module, python module is available only in PySide6.
    _QtModuleDef("QtStateMachine", shared_lib="StateMachine", bindings=["PySide6"]),

    # *** qt/qtsensors ***
    _QtModuleDef("QtSensors", shared_lib="Sensors", plugins=["sensors", "sensorgestures"]),

    # *** qt/qtserialport ***
    _QtModuleDef("QtSerialPort", shared_lib="SerialPort", translations=["qtserialport"]),

    # *** qt/qtscript ***
    # Qt5-only Qt module, python module is available only in PySide2. PyQt5 wheels do not seem to ship the corresponding
    # Qt modules (shared libs) at all.
    _QtModuleDef("QtScript", shared_lib="Script", translations=["qtscript"], plugins=["script"], bindings=["PySide2"]),
    _QtModuleDef("QtScriptTools", shared_lib="ScriptTools", bindings=["PySide2"]),

    # *** qt/qtserialbus ***
    # No python module; shared library -> plugins association entry.
    # PySide6 6.5.0 introduced python module.
    _QtModuleDef(None, shared_lib="SerialBus", plugins=["canbus"], bindings=["!PySide6"]),
    _QtModuleDef("QtSerialBus", shared_lib="SerialBus", plugins=["canbus"], bindings=["PySide6"]),

    # *** qt/qtsvg ***
    _QtModuleDef("QtSvg", shared_lib="Svg"),
    # Qt6-only Qt module.
    _QtModuleDef("QtSvgWidgets", shared_lib="SvgWidgets", bindings=["PySide6", "PyQt6"]),

    # *** qt/qtspeech ***
    _QtModuleDef("QtTextToSpeech", shared_lib="TextToSpeech", plugins=["texttospeech"]),

    # *** qt/qttools ***
    # QtDesigner python module is available in all bindings but PySide2.
    _QtModuleDef(None, shared_lib="Designer", plugins=["designer"], bindings=["PySide2"]),
    _QtModuleDef(
        "QtDesigner", shared_lib="Designer", translations=["designer"], plugins=["designer"], bindings=["!PySide2"]
    ),
    _QtModuleDef("QtHelp", shared_lib="Help", translations=["qt_help"]),
    # Python module is available only in PySide bindings.
    _QtModuleDef("QtUiTools", shared_lib="UiTools", bindings=["PySide*"]),

    # *** qt/qtvirtualkeyboard ***
    # No python module; shared library -> plugins association entry.
    _QtModuleDef(None, shared_lib="VirtualKeyboard", plugins=["virtualkeyboard"]),

    # *** qt/qtwebchannel ***
    _QtModuleDef("QtWebChannel", shared_lib="WebChannel"),

    # *** qt/qtwebengine ***
    # QtWebEngine is Qt5-only module (replaced by QtWebEngineQuick in Qt6).
    _QtModuleDef("QtWebEngine", shared_lib="WebEngine", bindings=["PySide2", "PyQt5"]),
    _QtModuleDef("QtWebEngineCore", shared_lib="WebEngineCore", translations=["qtwebengine"]),
    # QtWebEngineQuick is Qt6-only module (replacement for QtWebEngine in Qt5).
    _QtModuleDef("QtWebEngineQuick", shared_lib="WebEngineQuick", bindings=["PySide6", "PyQt6"]),
    _QtModuleDef("QtWebEngineWidgets", shared_lib="WebEngineWidgets"),
    # QtPdf and QtPdfWidgets have python module available in PySide6 and PyQt6 >= 6.4.0.
    _QtModuleDef("QtPdf", shared_lib="Pdf", bindings=["PySide6", "PyQt6"]),
    _QtModuleDef("QtPdfWidgets", shared_lib="PdfWidgets", bindings=["PySide6", "PyQt6"]),

    # *** qt/qtwebsockets ***
    _QtModuleDef("QtWebSockets", shared_lib="WebSockets", translations=["qtwebsockets"]),

    # *** qt/qtwebview ***
    # No python module; shared library -> plugins association entry.
    _QtModuleDef(None, shared_lib="WebView", plugins=["webview"]),

    # *** qt/qtwinextras ***
    # Qt5-only Qt module.
    _QtModuleDef("QtWinExtras", shared_lib="WinExtras", bindings=["PySide2", "PyQt5"]),

    # *** qt/qtx11extras ***
    # Qt5-only Qt module.
    _QtModuleDef("QtX11Extras", shared_lib="X11Extras", bindings=["PySide2", "PyQt5"]),

    # *** qt/qtxmlpatterns ***
    # Qt5-only Qt module.
    _QtModuleDef(
        "QtXmlPatterns", shared_lib="XmlPatterns", translations=["qtxmlpatterns"], bindings=["PySide2", "PyQt5"]
    ),

    # *** qscintilla ***
    # Python module is available only in PyQt bindings. No associated shared library.
    _QtModuleDef("Qsci", translations=["qscintilla"], bindings=["PyQt*"]),
)


# Helpers for turning Qt namespace specifiers, such as "!PySide2" or "PyQt*", into set of applicable
# namespaces.
def process_namespace_strings(namespaces):
    """"Process list of Qt namespace specifier strings into set of namespaces."""
    bindings = set()
    for namespace in namespaces:
        bindings |= _process_namespace_string(namespace)
    return bindings


def _process_namespace_string(namespace):
    """Expand a Qt namespace specifier string into set of namespaces."""
    if namespace.startswith("!"):
        bindings = _process_namespace_string(namespace[1:])
        return ALL_QT_BINDINGS - bindings
    else:
        if namespace == "PySide*":
            return {"PySide2", "PySide6"}
        elif namespace == "PyQt*":
            return {"PyQt5", "PyQt6"}
        elif namespace in ALL_QT_BINDINGS:
            return {namespace}
        else:
            raise ValueError(f"Invalid Qt namespace specifier: {namespace}!")



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, shutil, platform
from waflib import Task, Utils
from waflib.TaskGen import taskgen_method, feature, after_method, before_method

app_info = '''
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist SYSTEM "file://localhost/System/Library/DTDs/PropertyList.dtd">
<plist version="0.9">
<dict>
	<key>CFBundlePackageType</key>
	<string>APPL</string>
	<key>CFBundleGetInfoString</key>
	<string>Created by Waf</string>
	<key>CFBundleSignature</key>
	<string>????</string>
	<key>NOTE</key>
	<string>THIS IS A GENERATED FILE, DO NOT MODIFY</string>
	<key>CFBundleExecutable</key>
	<string>{app_name}</string>
</dict>
</plist>
'''


@feature('c', 'cxx')
def set_macosx_deployment_target(self):
    if self.env.MACOSX_DEPLOYMENT_TARGET:
        os.environ['MACOSX_DEPLOYMENT_TARGET'] = self.env.MACOSX_DEPLOYMENT_TARGET
    elif 'MACOSX_DEPLOYMENT_TARGET' not in os.environ:
        if Utils.unversioned_sys_platform() == 'darwin':
            os.environ['MACOSX_DEPLOYMENT_TARGET'] = '.'.join(platform.mac_ver()[0].split('.')[:2])


@taskgen_method
def create_bundle_dirs(self, name, out):
    dir = out.parent.find_or_declare(name)
    dir.mkdir()
    macos = dir.find_or_declare(['Contents', 'MacOS'])
    macos.mkdir()
    return dir


def bundle_name_for_output(out):
    name = out.name
    k = name.rfind('.')
    if k >= 0:
        name = name[:k] + '.app'
    else:
        name = name + '.app'
    return name


@feature('cprogram', 'cxxprogram')
@after_method('apply_link')
def create_task_macapp(self):
    if self.env.MACAPP or getattr(self, 'mac_app', False):
        out = self.link_task.outputs[0]
        name = bundle_name_for_output(out)
        dir = self.create_bundle_dirs(name, out)
        n1 = dir.find_or_declare(['Contents', 'MacOS', out.name])
        self.apptask = self.create_task('macapp', self.link_task.outputs, n1)
        inst_to = getattr(self, 'install_path', '/Applications') + '/%s/Contents/MacOS/' % name
        self.add_install_files(install_to=inst_to, install_from=n1, chmod=Utils.O755)
        if getattr(self, 'mac_files', None):
            mac_files_root = getattr(self, 'mac_files_root', None)
            if isinstance(mac_files_root, str):
                mac_files_root = self.path.find_node(mac_files_root)
                if not mac_files_root:
                    self.bld.fatal('Invalid mac_files_root %r' % self.mac_files_root)
            res_dir = n1.parent.parent.make_node('Resources')
            inst_to = getattr(self, 'install_path', '/Applications') + '/%s/Resources' % name
            for node in self.to_nodes(self.mac_files):
                relpath = node.path_from(mac_files_root or node.parent)
                self.create_task('macapp', node, res_dir.make_node(relpath))
                self.add_install_as(install_to=os.path.join(inst_to, relpath), install_from=node)
        if getattr(self.bld, 'is_install', None):
            self.install_task.hasrun = Task.SKIP_ME


@feature('cprogram', 'cxxprogram')
@after_method('apply_link')
def create_task_macplist(self):
    if self.env.MACAPP or getattr(self, 'mac_app', False):
        out = self.link_task.outputs[0]
        name = bundle_name_for_output(out)
        dir = self.create_bundle_dirs(name, out)
        n1 = dir.find_or_declare(['Contents', 'Info.plist'])
        self.plisttask = plisttask = self.create_task('macplist', [], n1)
        plisttask.context = {'app_name': self.link_task.outputs[0].name, 'env': self.env}
        plist_ctx = getattr(self, 'plist_context', None)
        if (plist_ctx):
            plisttask.context.update(plist_ctx)
        if getattr(self, 'mac_plist', False):
            node = self.path.find_resource(self.mac_plist)
            if node:
                plisttask.inputs.append(node)
            else:
                plisttask.code = self.mac_plist
        else:
            plisttask.code = app_info
        inst_to = getattr(self, 'install_path', '/Applications') + '/%s/Contents/' % name
        self.add_install_files(install_to=inst_to, install_from=n1)


@feature('cshlib', 'cxxshlib')
@before_method('apply_link', 'propagate_uselib_vars')
def apply_bundle(self):
    if self.env.MACBUNDLE or getattr(self, 'mac_bundle', False):
        self.env.LINKFLAGS_cshlib = self.env.LINKFLAGS_cxxshlib = []
        self.env.cshlib_PATTERN = self.env.cxxshlib_PATTERN = self.env.macbundle_PATTERN
        use = self.use = self.to_list(getattr(self, 'use', []))
        if not 'MACBUNDLE' in use:
            use.append('MACBUNDLE')


app_dirs = ['Contents', 'Contents/MacOS', 'Contents/Resources']


class macapp(Task.Task):
    color = 'PINK'

    def run(self):
        self.outputs[0].parent.mkdir()
        shutil.copy2(self.inputs[0].srcpath(), self.outputs[0].abspath())


class macplist(Task.Task):
    color = 'PINK'
    ext_in = ['.bin']

    def run(self):
        if getattr(self, 'code', None):
            txt = self.code
        else:
            txt = self.inputs[0].read()
        context = getattr(self, 'context', {})
        txt = txt.format(**context)
        self.outputs[0].write(txt)



from pkg import toplevel_existing
from pkg import toplevel_nonexisting

class MyClass:
    from pkg import toplevel_class_existing
    from pkg import toplevel_class_nonexisting

if a == b:
    from pkg import toplevel_conditional_existing
    from pkg import toplevel_conditional_nonexisting

    try:
        from pkg import toplevel_conditional_import_existing, toplevel_conditional_import_nonexisting
    except:
        from pkg import toplevel_conditional_import2_existing
        from pkg import toplevel_conditional_import2_nonexisting

try:
    from pkg import toplevel_import_existing, toplevel_import_nonexisting
except:
    from pkg import toplevel_import2_existing
    from pkg import toplevel_import2_nonexisting

def function():
    from pkg import function_existing, function_nonexisting

    class MyClass:
        from pkg import function_class_existing, function_class_nonexisting

    if a == b:
        from pkg import function_conditional_existing
        from pkg import function_conditional_nonexisting

        try:
            from pkg import function_conditional_import_existing
            from pkg import function_conditional_import_nonexisting
        except:
            from pkg import function_conditional_import2_existing
            from pkg import function_conditional_import2_nonexisting

    try:
        from pkg import function_import_existing
        from pkg import function_import_nonexisting
    except:
        from pkg import function_import2_existing
        from pkg import function_import2_nonexisting



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, re
from waflib import Task, Utils, Node, Errors, Logs
from waflib.TaskGen import after_method, before_method, feature, taskgen_method, extension
from waflib.Tools import c_aliases, c_preproc, c_config, c_osx, c_tests
from waflib.Configure import conf

SYSTEM_LIB_PATHS = ['/usr/lib64', '/usr/lib', '/usr/local/lib64', '/usr/local/lib']
USELIB_VARS = Utils.defaultdict(set)
USELIB_VARS['c'] = set(['INCLUDES', 'FRAMEWORKPATH', 'DEFINES', 'CPPFLAGS', 'CCDEPS', 'CFLAGS', 'ARCH'])
USELIB_VARS['cxx'] = set(['INCLUDES', 'FRAMEWORKPATH', 'DEFINES', 'CPPFLAGS', 'CXXDEPS', 'CXXFLAGS', 'ARCH'])
USELIB_VARS['d'] = set(['INCLUDES', 'DFLAGS'])
USELIB_VARS['includes'] = set(['INCLUDES', 'FRAMEWORKPATH', 'ARCH'])
USELIB_VARS['cprogram'] = USELIB_VARS['cxxprogram'] = set([
    'LIB', 'STLIB', 'LIBPATH', 'STLIBPATH', 'LINKFLAGS', 'RPATH', 'LINKDEPS', 'FRAMEWORK', 'FRAMEWORKPATH', 'ARCH',
    'LDFLAGS'
])
USELIB_VARS['cshlib'] = USELIB_VARS['cxxshlib'] = set([
    'LIB', 'STLIB', 'LIBPATH', 'STLIBPATH', 'LINKFLAGS', 'RPATH', 'LINKDEPS', 'FRAMEWORK', 'FRAMEWORKPATH', 'ARCH',
    'LDFLAGS'
])
USELIB_VARS['cstlib'] = USELIB_VARS['cxxstlib'] = set(['ARFLAGS', 'LINKDEPS'])
USELIB_VARS['dprogram'] = set(['LIB', 'STLIB', 'LIBPATH', 'STLIBPATH', 'LINKFLAGS', 'RPATH', 'LINKDEPS'])
USELIB_VARS['dshlib'] = set(['LIB', 'STLIB', 'LIBPATH', 'STLIBPATH', 'LINKFLAGS', 'RPATH', 'LINKDEPS'])
USELIB_VARS['dstlib'] = set(['ARFLAGS', 'LINKDEPS'])
USELIB_VARS['asm'] = set(['ASFLAGS'])


@taskgen_method
def create_compiled_task(self, name, node):
    out = '%s.%d.o' % (node.name, self.idx)
    task = self.create_task(name, node, node.parent.find_or_declare(out))
    try:
        self.compiled_tasks.append(task)
    except AttributeError:
        self.compiled_tasks = [task]
    return task


@taskgen_method
def to_incnodes(self, inlst):
    lst = []
    seen = set()
    for x in self.to_list(inlst):
        if x in seen or not x:
            continue
        seen.add(x)
        if isinstance(x, Node.Node):
            lst.append(x)
        else:
            if os.path.isabs(x):
                lst.append(self.bld.root.make_node(x) or x)
            else:
                if x[0] == '#':
                    p = self.bld.bldnode.make_node(x[1:])
                    v = self.bld.srcnode.make_node(x[1:])
                else:
                    p = self.path.get_bld().make_node(x)
                    v = self.path.make_node(x)
                if p.is_child_of(self.bld.bldnode):
                    p.mkdir()
                lst.append(p)
                lst.append(v)
    return lst


@feature('c', 'cxx', 'd', 'asm', 'fc', 'includes')
@after_method('propagate_uselib_vars', 'process_source')
def apply_incpaths(self):
    lst = self.to_incnodes(self.to_list(getattr(self, 'includes', [])) + self.env.INCLUDES)
    self.includes_nodes = lst
    cwd = self.get_cwd()
    if Utils.is_win32:
        # Visual Studio limitations
        self.env.INCPATHS = [x.path_from(cwd) if x.is_child_of(self.bld.srcnode) else x.abspath() for x in lst]
    else:
        self.env.INCPATHS = [x.path_from(cwd) for x in lst]


class link_task(Task.Task):
    color = 'YELLOW'
    weight = 3
    inst_to = None
    chmod = Utils.O755

    def add_target(self, target):
        if isinstance(target, str):
            base = self.generator.path
            if target.startswith('#'):
                target = target[1:]
                base = self.generator.bld.bldnode
            pattern = self.env[self.__class__.__name__ + '_PATTERN']
            if not pattern:
                pattern = '%s'
            folder, name = os.path.split(target)
            if self.__class__.__name__.find('shlib') > 0 and getattr(self.generator, 'vnum', None):
                nums = self.generator.vnum.split('.')
                if self.env.DEST_BINFMT == 'pe':
                    name = name + '-' + nums[0]
                elif self.env.DEST_OS == 'openbsd':
                    pattern = '%s.%s' % (pattern, nums[0])
                    if len(nums) >= 2:
                        pattern += '.%s' % nums[1]
            if folder:
                tmp = folder + os.sep + pattern % name
            else:
                tmp = pattern % name
            target = base.find_or_declare(tmp)
        self.set_outputs(target)

    def exec_command(self, *k, **kw):
        ret = super(link_task, self).exec_command(*k, **kw)
        if not ret and self.env.DO_MANIFEST:
            ret = self.exec_mf()
        return ret

    def exec_mf(self):
        if not self.env.MT:
            return 0
        manifest = None
        for out_node in self.outputs:
            if out_node.name.endswith('.manifest'):
                manifest = out_node.abspath()
                break
        else:
            return 0
        mode = ''
        for x in Utils.to_list(self.generator.features):
            if x in ('cprogram', 'cxxprogram', 'fcprogram', 'fcprogram_test'):
                mode = 1
            elif x in ('cshlib', 'cxxshlib', 'fcshlib'):
                mode = 2
        Logs.debug('msvc: embedding manifest in mode %r', mode)
        lst = [] + self.env.MT
        lst.extend(Utils.to_list(self.env.MTFLAGS))
        lst.extend(['-manifest', manifest])
        lst.append('-outputresource:%s;%s' % (self.outputs[0].abspath(), mode))
        return super(link_task, self).exec_command(lst)


class stlink_task(link_task):
    run_str = '${AR} ${ARFLAGS} ${AR_TGT_F}${TGT} ${AR_SRC_F}${SRC}'
    chmod = Utils.O644


def rm_tgt(cls):
    old = cls.run

    def wrap(self):
        try:
            os.remove(self.outputs[0].abspath())
        except OSError:
            pass
        return old(self)

    setattr(cls, 'run', wrap)


rm_tgt(stlink_task)


@feature('skip_stlib_link_deps')
@before_method('process_use')
def apply_skip_stlib_link_deps(self):
    self.env.SKIP_STLIB_LINK_DEPS = True


@feature('c', 'cxx', 'd', 'fc', 'asm')
@after_method('process_source')
def apply_link(self):
    for x in self.features:
        if x == 'cprogram' and 'cxx' in self.features:
            x = 'cxxprogram'
        elif x == 'cshlib' and 'cxx' in self.features:
            x = 'cxxshlib'
        if x in Task.classes:
            if issubclass(Task.classes[x], link_task):
                link = x
                break
    else:
        return
    objs = [t.outputs[0] for t in getattr(self, 'compiled_tasks', [])]
    self.link_task = self.create_task(link, objs)
    self.link_task.add_target(self.target)
    try:
        inst_to = self.install_path
    except AttributeError:
        inst_to = self.link_task.inst_to
    if inst_to:
        self.install_task = self.add_install_files(
            install_to=inst_to, install_from=self.link_task.outputs[:], chmod=self.link_task.chmod, task=self.link_task
        )


@taskgen_method
def use_rec(self, name, **kw):
    if name in self.tmp_use_not or name in self.tmp_use_seen:
        return
    try:
        y = self.bld.get_tgen_by_name(name)
    except Errors.WafError:
        self.uselib.append(name)
        self.tmp_use_not.add(name)
        return
    self.tmp_use_seen.append(name)
    y.post()
    y.tmp_use_objects = objects = kw.get('objects', True)
    y.tmp_use_stlib = stlib = kw.get('stlib', True)
    try:
        link_task = y.link_task
    except AttributeError:
        y.tmp_use_var = ''
    else:
        objects = False
        if not isinstance(link_task, stlink_task):
            stlib = False
            y.tmp_use_var = 'LIB'
        else:
            y.tmp_use_var = 'STLIB'
    p = self.tmp_use_prec
    for x in self.to_list(getattr(y, 'use', [])):
        if self.env["STLIB_" + x]:
            continue
        try:
            p[x].append(name)
        except KeyError:
            p[x] = [name]
        self.use_rec(x, objects=objects, stlib=stlib)


@feature('c', 'cxx', 'd', 'use', 'fc')
@before_method('apply_incpaths', 'propagate_uselib_vars')
@after_method('apply_link', 'process_source')
def process_use(self):
    use_not = self.tmp_use_not = set()
    self.tmp_use_seen = []
    use_prec = self.tmp_use_prec = {}
    self.uselib = self.to_list(getattr(self, 'uselib', []))
    self.includes = self.to_list(getattr(self, 'includes', []))
    names = self.to_list(getattr(self, 'use', []))
    for x in names:
        self.use_rec(x)
    for x in use_not:
        if x in use_prec:
            del use_prec[x]
    out = self.tmp_use_sorted = []
    tmp = []
    for x in self.tmp_use_seen:
        for k in use_prec.values():
            if x in k:
                break
        else:
            tmp.append(x)
    while tmp:
        e = tmp.pop()
        out.append(e)
        try:
            nlst = use_prec[e]
        except KeyError:
            pass
        else:
            del use_prec[e]
            for x in nlst:
                for y in use_prec:
                    if x in use_prec[y]:
                        break
                else:
                    tmp.append(x)
    if use_prec:
        raise Errors.WafError('Cycle detected in the use processing %r' % use_prec)
    out.reverse()
    link_task = getattr(self, 'link_task', None)
    for x in out:
        y = self.bld.get_tgen_by_name(x)
        var = y.tmp_use_var
        if var and link_task:
            if self.env.SKIP_STLIB_LINK_DEPS and isinstance(link_task, stlink_task):
                pass
            elif var == 'LIB' or y.tmp_use_stlib or x in names:
                self.env.append_value(var, [y.target[y.target.rfind(os.sep) + 1:]])
                self.link_task.dep_nodes.extend(y.link_task.outputs)
                tmp_path = y.link_task.outputs[0].parent.path_from(self.get_cwd())
                self.env.append_unique(var + 'PATH', [tmp_path])
        else:
            if y.tmp_use_objects:
                self.add_objects_from_tgen(y)
        if getattr(y, 'export_includes', None):
            self.includes = self.includes + y.to_incnodes(y.export_includes)
        if getattr(y, 'export_defines', None):
            self.env.append_value('DEFINES', self.to_list(y.export_defines))
    for x in names:
        try:
            y = self.bld.get_tgen_by_name(x)
        except Errors.WafError:
            if not self.env['STLIB_' + x] and not x in self.uselib:
                self.uselib.append(x)
        else:
            for k in self.to_list(getattr(y, 'use', [])):
                if not self.env['STLIB_' + k] and not k in self.uselib:
                    self.uselib.append(k)


@taskgen_method
def accept_node_to_link(self, node):
    return not node.name.endswith('.pdb')


@taskgen_method
def add_objects_from_tgen(self, tg):
    try:
        link_task = self.link_task
    except AttributeError:
        pass
    else:
        for tsk in getattr(tg, 'compiled_tasks', []):
            for x in tsk.outputs:
                if self.accept_node_to_link(x):
                    link_task.inputs.append(x)


@taskgen_method
def get_uselib_vars(self):
    _vars = set()
    for x in self.features:
        if x in USELIB_VARS:
            _vars |= USELIB_VARS[x]
    return _vars


@feature('c', 'cxx', 'd', 'fc', 'javac', 'cs', 'uselib', 'asm')
@after_method('process_use')
def propagate_uselib_vars(self):
    _vars = self.get_uselib_vars()
    env = self.env
    app = env.append_value
    feature_uselib = self.features + self.to_list(getattr(self, 'uselib', []))
    for var in _vars:
        y = var.lower()
        val = getattr(self, y, [])
        if val:
            app(var, self.to_list(val))
        for x in feature_uselib:
            val = env['%s_%s' % (var, x)]
            if val:
                app(var, val)


@feature('cshlib', 'cxxshlib', 'fcshlib')
@after_method('apply_link')
def apply_implib(self):
    if not self.env.DEST_BINFMT == 'pe':
        return
    dll = self.link_task.outputs[0]
    if isinstance(self.target, Node.Node):
        name = self.target.name
    else:
        name = os.path.split(self.target)[1]
    implib = self.env.implib_PATTERN % name
    implib = dll.parent.find_or_declare(implib)
    self.env.append_value('LINKFLAGS', self.env.IMPLIB_ST % implib.bldpath())
    self.link_task.outputs.append(implib)
    if getattr(self, 'defs', None) and self.env.DEST_BINFMT == 'pe':
        node = self.path.find_resource(self.defs)
        if not node:
            raise Errors.WafError('invalid def file %r' % self.defs)
        if self.env.def_PATTERN:
            self.env.append_value('LINKFLAGS', self.env.def_PATTERN % node.path_from(self.get_cwd()))
            self.link_task.dep_nodes.append(node)
        else:
            self.link_task.inputs.append(node)
    if getattr(self, 'install_task', None):
        try:
            inst_to = self.install_path_implib
        except AttributeError:
            try:
                inst_to = self.install_path
            except AttributeError:
                inst_to = '${IMPLIBDIR}'
                self.install_task.install_to = '${BINDIR}'
                if not self.env.IMPLIBDIR:
                    self.env.IMPLIBDIR = self.env.LIBDIR
        self.implib_install_task = self.add_install_files(
            install_to=inst_to, install_from=implib, chmod=self.link_task.chmod, task=self.link_task
        )


re_vnum = re.compile('^([1-9]\\d*|0)([.]([1-9]\\d*|0)){0,2}?$')


@feature('cshlib', 'cxxshlib', 'dshlib', 'fcshlib', 'vnum')
@after_method('apply_link', 'propagate_uselib_vars')
def apply_vnum(self):
    if not getattr(self, 'vnum', '') or os.name != 'posix' or self.env.DEST_BINFMT not in ('elf', 'mac-o'):
        return
    link = self.link_task
    if not re_vnum.match(self.vnum):
        raise Errors.WafError('Invalid vnum %r for target %r' % (self.vnum, getattr(self, 'name', self)))
    nums = self.vnum.split('.')
    node = link.outputs[0]
    cnum = getattr(self, 'cnum', str(nums[0]))
    cnums = cnum.split('.')
    if len(cnums) > len(nums) or nums[0:len(cnums)] != cnums:
        raise Errors.WafError('invalid compatibility version %s' % cnum)
    libname = node.name
    if libname.endswith('.dylib'):
        name3 = libname.replace('.dylib', '.%s.dylib' % self.vnum)
        name2 = libname.replace('.dylib', '.%s.dylib' % cnum)
    else:
        name3 = libname + '.' + self.vnum
        name2 = libname + '.' + cnum
    if self.env.SONAME_ST:
        v = self.env.SONAME_ST % name2
        self.env.append_value('LINKFLAGS', v.split())
    if self.env.DEST_OS != 'openbsd':
        outs = [node.parent.make_node(name3)]
        if name2 != name3:
            outs.append(node.parent.make_node(name2))
        self.create_task('vnum', node, outs)
    if getattr(self, 'install_task', None):
        self.install_task.hasrun = Task.SKIPPED
        self.install_task.no_errcheck_out = True
        path = self.install_task.install_to
        if self.env.DEST_OS == 'openbsd':
            libname = self.link_task.outputs[0].name
            t1 = self.add_install_as(
                install_to='%s/%s' % (path, libname), install_from=node, chmod=self.link_task.chmod
            )
            self.vnum_install_task = (t1,)
        else:
            t1 = self.add_install_as(install_to=path + os.sep + name3, install_from=node, chmod=self.link_task.chmod)
            t3 = self.add_symlink_as(install_to=path + os.sep + libname, install_from=name3)
            if name2 != name3:
                t2 = self.add_symlink_as(install_to=path + os.sep + name2, install_from=name3)
                self.vnum_install_task = (t1, t2, t3)
            else:
                self.vnum_install_task = (t1, t3)
    if '-dynamiclib' in self.env.LINKFLAGS:
        try:
            inst_to = self.install_path
        except AttributeError:
            inst_to = self.link_task.inst_to
        if inst_to:
            p = Utils.subst_vars(inst_to, self.env)
            path = os.path.join(p, name2)
            self.env.append_value('LINKFLAGS', ['-install_name', path])
            self.env.append_value('LINKFLAGS', '-Wl,-compatibility_version,%s' % cnum)
            self.env.append_value('LINKFLAGS', '-Wl,-current_version,%s' % self.vnum)


class vnum(Task.Task):
    color = 'CYAN'
    ext_in = ['.bin']

    def keyword(self):
        return 'Symlinking'

    def run(self):
        for x in self.outputs:
            path = x.abspath()
            try:
                os.remove(path)
            except OSError:
                pass
            try:
                os.symlink(self.inputs[0].name, path)
            except OSError:
                return 1


class fake_shlib(link_task):
    def runnable_status(self):
        for t in self.run_after:
            if not t.hasrun:
                return Task.ASK_LATER
        return Task.SKIP_ME


class fake_stlib(stlink_task):
    def runnable_status(self):
        for t in self.run_after:
            if not t.hasrun:
                return Task.ASK_LATER
        return Task.SKIP_ME


@conf
def read_shlib(self, name, paths=[], export_includes=[], export_defines=[]):
    return self(
        name=name,
        features='fake_lib',
        lib_paths=paths,
        lib_type='shlib',
        export_includes=export_includes,
        export_defines=export_defines
    )


@conf
def read_stlib(self, name, paths=[], export_includes=[], export_defines=[]):
    return self(
        name=name,
        features='fake_lib',
        lib_paths=paths,
        lib_type='stlib',
        export_includes=export_includes,
        export_defines=export_defines
    )


lib_patterns = {
    'shlib': ['lib%s.so', '%s.so', 'lib%s.dylib', 'lib%s.dll', '%s.dll'],
    'stlib': ['lib%s.a', '%s.a', 'lib%s.dll', '%s.dll', 'lib%s.lib', '%s.lib'],
}


@feature('fake_lib')
def process_lib(self):
    node = None
    names = [x % self.name for x in lib_patterns[self.lib_type]]
    for x in self.lib_paths + [self.path] + SYSTEM_LIB_PATHS:
        if not isinstance(x, Node.Node):
            x = self.bld.root.find_node(x) or self.path.find_node(x)
            if not x:
                continue
        for y in names:
            node = x.find_node(y)
            if node:
                try:
                    Utils.h_file(node.abspath())
                except EnvironmentError:
                    raise ValueError('Could not read %r' % y)
                break
        else:
            continue
        break
    else:
        raise Errors.WafError('could not find library %r' % self.name)
    self.link_task = self.create_task('fake_%s' % self.lib_type, [], [node])
    self.target = self.name


class fake_o(Task.Task):
    def runnable_status(self):
        return Task.SKIP_ME


@extension('.o', '.obj')
def add_those_o_files(self, node):
    tsk = self.create_task('fake_o', [], node)
    try:
        self.compiled_tasks.append(tsk)
    except AttributeError:
        self.compiled_tasks = [tsk]


@feature('fake_obj')
@before_method('process_source')
def process_objs(self):
    for node in self.to_nodes(self.source):
        self.add_those_o_files(node)
    self.source = []


@conf
def read_object(self, obj):
    if not isinstance(obj, self.path.__class__):
        obj = self.path.find_resource(obj)
    return self(features='fake_obj', source=obj, name=obj.name)


@feature('cxxprogram', 'cprogram')
@after_method('apply_link', 'process_use')
def set_full_paths_hpux(self):
    if self.env.DEST_OS != 'hp-ux':
        return
    base = self.bld.bldnode.abspath()
    for var in ['LIBPATH', 'STLIBPATH']:
        lst = []
        for x in self.env[var]:
            if x.startswith('/'):
                lst.append(x)
            else:
                lst.append(os.path.normpath(os.path.join(base, x)))
        self.env[var] = lst



from setuptools import setup

setup(
    name='pyi_example_package',
    setup_requires="setuptools >= 40.0.0",
    packages=['pyi_example_package'],
    author='PyInstaller development team',
    entry_points={'pyinstaller40': ['hooks = pyi_example_package.__init__:get_hook_dirs']}
)



import unittest

import os, py_compile, sys

from PyInstaller.lib.modulegraph import modulegraph

class CompiledModuleTests(unittest.TestCase):
    def setUp(self):
        self.base_dir = os.path.join(
                os.path.dirname(__file__),
                'testpkg-compiled')
        self.compiled_dir = os.path.join(
                self.base_dir, 'compiled')
        self.source_dir = os.path.join(
                self.base_dir, 'source')

        for fn in os.listdir(self.source_dir):
            if not fn.endswith('.py'): continue

            py_compile.compile(
                os.path.join(self.source_dir, fn),
                os.path.join(self.compiled_dir, fn + 'c'))

    def tearDown(self):
        for fn in os.listdir(self.compiled_dir):
            if fn.endswith('.pyc'):
                os.unlink(os.path.join(self.compiled_dir, fn))

    def testCompiledModules(self):
        mf = modulegraph.ModuleGraph(path=[self.compiled_dir] + sys.path)
        #self.mf.debug = 999
        mf.add_script(os.path.join(self.compiled_dir, 'script.py'))

        o = mf.find_node('mod1')
        self.assertIsInstance(o, modulegraph.CompiledModule)
        self.assertEqual(o._global_attr_names, { 'mod2', 'mod3', 'foo' })
        self.assertEqual(o._starimported_ignored_module_names, set())

        o = mf.find_node('mod2')
        self.assertIsInstance(o, modulegraph.CompiledModule)
        self.assertEqual(o._global_attr_names,
                         { 'mod1', 'sys', 'testme', 'bar' })
        self.assertEqual(o._starimported_ignored_module_names, set())

        o = mf.find_node('mod3')
        self.assertIsInstance(o, modulegraph.CompiledModule)
        self.assertEqual(o._global_attr_names, { 'os', 'path'})
        self.assertEqual(o._starimported_ignored_module_names, set())

        o = mf.find_node('mod4')
        other = mf.find_node('zipfile')
        self.assertIsInstance(o, modulegraph.CompiledModule)
        self.assertEqual(o._global_attr_names, other._global_attr_names)
        self.assertEqual(o._starimported_ignored_module_names, {'math'})

        o = mf.find_node('mod5')
        self.assertIs(o, None)


if __name__ == '__main__':
    #import profile
    #profile.run('unittest.main()', sort=2)
    unittest.main()



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

from waflib.Configure import conf


@conf
def find_ar(conf):
    conf.load('ar')


def configure(conf):
    conf.find_program('ar', var='AR')
    conf.add_os_flags('ARFLAGS')
    if not conf.env.ARFLAGS:
        conf.env.ARFLAGS = ['rcs']



#-----------------------------------------------------------------------------
# Copyright (c) 2025, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# This top-level namespace package might be provided by setuptools >= 71.0.0, which makes its vendored dependencies
# public by appending path to its `setuptools._vendored` directory to `sys.path`. The following shared
# pre-safe-import-module hook implementation checks whether this is the case, and, depending on situation, either
# sets up aliases to prevent duplicate collection, or extends the search paths.
from PyInstaller.utils.hooks.setuptools import pre_safe_import_module_for_top_level_namespace_packages \
    as pre_safe_import_module  # noqa: F401



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

# PEP-302 import hooks specification contain section 'Optional Extensions to the Importer Protocol'
#
# This section is meant to be optional but the reality is different. Some Python modules (e.g. Flask) depends on
# implementation of these optional functions:
#
#   loader.is_package(fullmodname)
#   loader.get_code(fullmodname)
#   loader.get_source(fullmodname)
#
# This test-cases test the return values of these functions for importers from pyimod02_importers module.

# Note: The modules need to be imported at the end of the resp. code. Otherwise the pkgutil-functions take a very
#       different branch (since the module is already in sys.modules) and what we want to test will not be tested.

import pytest

from PyInstaller import compat

# All tests here require `pkgutil.get_loader()`, which was deprecated in python 3.12 and removed in 3.14.
pytestmark = pytest.mark.skipif(compat.is_py314, reason="pkgutil.get_loader() was removed in python >= 3.14.")


def test_pep302_loader_builtin(pyi_builder):
    pyi_builder.test_source(
        """
        mod = 'sys'
        import pkgutil
        ldr = pkgutil.get_loader(mod)
        assert ldr
        assert ldr.is_package(mod) == False
        assert ldr.get_code(mod) is None
        assert ldr.get_source(mod) is None
        """
    )


def test_pep302_loader_frozen_module(pyi_builder):
    pyi_builder.test_source(
        """
        mod = 'compileall'
        import pkgutil
        ldr = pkgutil.get_loader(mod)
        assert ldr
        assert ldr.is_package(mod) == False
        assert ldr.get_code(mod) is not None
        assert ldr.get_source(mod) is None
        # Import at the very end, just to get the module frozen.
        import compileall
        """
    )


def test_pep302_loader_frozen_package(pyi_builder):
    pyi_builder.test_source(
        """
        mod = 'json'
        import pkgutil
        ldr = pkgutil.get_loader(mod)
        assert ldr
        assert ldr.is_package(mod) == True
        assert ldr.get_code(mod) is not None
        assert ldr.get_source(mod) is None
        # Import at the very end, just to get the module frozen.
        import json
        """
    )


def test_pep302_loader_frozen_submodule(pyi_builder):
    pyi_builder.test_source(
        """
        mod = 'json.encoder'
        import pkgutil
        ldr = pkgutil.get_loader(mod)
        assert ldr
        assert ldr.is_package(mod) == False
        assert ldr.get_code(mod) is not None
        assert ldr.get_source(mod) is None
        # Import at the very end, just to get the module frozen.
        import json.encoder
        """
    )


def test_pep302_loader_cextension(pyi_builder):
    pyi_builder.test_source(
        """
        mod = '_sqlite3'
        import pkgutil
        ldr = pkgutil.get_loader(mod)
        assert ldr
        assert ldr.is_package(mod) == False
        assert ldr.get_code(mod) is None
        assert ldr.get_source(mod) is None
        # Import at the very end, just to get the module frozen.
        import sqlite3
        """
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2021-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

import os
import pathlib
import sys
import subprocess
import json

import pytest

from PyInstaller import compat
from PyInstaller.utils.tests import onefile_only

# Directory with testing modules used in some tests.
_MODULES_DIR = pathlib.Path(__file__).parent / 'modules'


# Test that in python 3.11 and later, sys._stdlib_dir is set and that python-frozen modules have __file__ attribute.
@pytest.mark.skipif(not compat.is_py311, reason="applicable only to python >= 3.11")
def test_frozen_stdlib_modules(pyi_builder, script_dir, tmp_path):
    test_script = 'pyi_frozen_stdlib_modules.py'
    ref_result_file = tmp_path / 'ref_results.txt'
    result_file = tmp_path / 'results.txt'

    # Run the test script unfrozen, to obtain reference results
    ret = compat.exec_python_rc(
        str(script_dir / test_script),
        str(ref_result_file),
    )
    assert ret == 0, "Unfrozen test script failed!"

    # Freeze and run the test script
    pyi_builder.test_script(
        test_script,
        app_args=[str(result_file)],
    )

    # Process the results
    def _normalize_module_path(module_path, stdlib_dir):
        if not module_path:
            return module_path
        module_path, ext = os.path.splitext(os.path.relpath(module_path, stdlib_dir))
        assert ext in ('.pyc', '.py')
        return module_path

    def _load_results(filename):
        # Read pprint-ed results
        with open(filename, 'r', encoding='utf-8') as fp:
            data = fp.read()
        data = eval(data)

        # First entry is sys._stdlib_dir
        stdlib_dir = data[0]

        results = []
        for name, file_attr, state_filename, state_origname in data[1:]:
            # Remove sys._stdlib_dir prefix from __file__ attribute and filename from __spec__.loader_state, and remove
            # the .py/.pyc suffix for easier comparison.
            results.append((
                name,
                _normalize_module_path(file_attr, stdlib_dir),
                _normalize_module_path(state_filename, stdlib_dir),
                state_origname,
            ))

        return results

    ref_results = _load_results(ref_result_file)
    results = _load_results(result_file)

    assert results == ref_results


# Test whether dis can disassemble the __main__ module, as per #5897.
def test_dis_main(pyi_builder):
    pyi_builder.test_source(
        """
        import dis
        import sys

        print(dis.dis(sys.modules["__main__"].__loader__.get_code("__main__")))
        """
    )


# Test that setting utf8 X-flag controls the PEP540 UTF-8 mode on all OSes, regardless of current locale setting.
@pytest.mark.parametrize('xflag,enabled', [("X utf8", True), ("X utf8=1", True), ("X utf8=0", False)])
def test_utf8_mode_xflag(xflag, enabled, pyi_builder):
    pyi_builder.test_source(
        f"""
        import sys
        print("sys.flags:", sys.flags)
        assert sys.flags.utf8_mode == {enabled}
        """,
        pyi_args=["--python-option", xflag]
    )


# Test that PEP540 UTF-8 mode is automatically enabled for C and POSIX locales (applicable only to macOS and linux).
@pytest.mark.linux
@pytest.mark.darwin
@pytest.mark.parametrize('locale', ['C', 'POSIX'])
def test_utf8_mode_locale(locale, pyi_builder, monkeypatch):
    monkeypatch.setenv('LC_CTYPE', locale)
    monkeypatch.setenv('LC_ALL', locale)  # Required by macOS CI; setting just LC_CTYPE is not enough.
    pyi_builder.test_source(
        """
        import sys
        print("sys.flags:", sys.flags)
        assert sys.flags.utf8_mode == 1
        """
    )


# Test that setting dev X-flag controls dev mode.
@pytest.mark.parametrize('xflag,enabled', [("X dev", True), ("X dev=1", True), ("X dev=0", False)])
def test_dev_mode_xflag(xflag, enabled, pyi_builder):
    pyi_builder.test_source(
        f"""
        import sys
        print("sys.flags:", sys.flags)
        assert sys.flags.dev_mode == {enabled}
        """,
        pyi_args=["--python-option", xflag]
    )


# Test that setting hash seed to zero via --python-option disables hash randomization.
def test_disable_hash_randomization(pyi_builder):
    pyi_builder.test_source(
        """
        import sys
        print("sys.flags:", sys.flags)
        assert sys.flags.hash_randomization == 0
        """,
        pyi_args=["--python-option", "hash_seed=0"]
    )


# Test that onefile cleanup does not remove contents of a directory that user symlinks into sys._MEIPASS (see #6074).
@onefile_only
def test_onefile_cleanup_symlinked_dir(pyi_builder, tmp_path):
    # Create output directory with five pre-existing files
    output_dir = tmp_path / 'output_dir'
    output_dir.mkdir()
    for idx in range(5):
        output_file = output_dir / f'preexisting-{idx}.txt'
        output_file.write_text(f"Pre-existing file #{idx}", encoding='utf-8')

    # Check if OS supports creation of symbolic links
    try:
        (tmp_path / 'testdir').symlink_to(output_dir)
    except OSError:
        pytest.skip("OS does not support (unprivileged) creation of symbolic links.")

    # Run the test program
    pyi_builder.test_source(
        """
        import sys
        import os

        # Output directory is passed via argv[1]; create symlink to it inside the _MEIPASS
        output_dir = os.path.join(sys._MEIPASS, 'output')
        os.symlink(sys.argv[1], output_dir)

        # Create five files
        for idx in range(5):
            output_file = os.path.join(output_dir, f'output-{idx}.txt')
            with open(output_file, 'w', encoding='utf-8') as fp:
                fp.write(f'Output file #{idx}')
        """,
        app_args=[output_dir]
    )

    # Output directory should contain all five pre-existing and five new files.
    for idx in range(5):
        output_file = output_dir / f'preexisting-{idx}.txt'
        assert output_file.is_file()
    for idx in range(5):
        output_file = output_dir / f'output-{idx}.txt'
        assert output_file.is_file()


# Test that single-file metadata (as commonly found in Debian/Ubuntu packages) is properly collected by copy_metadata().
def test_single_file_metadata(pyi_builder):
    # Add directory containing the my-test-package metadata to search path
    extra_path = _MODULES_DIR / "pyi_single_file_metadata"

    pyi_builder.test_source(
        """
        import pkg_resources

        # The pkg_resources.get_distribution() call automatically triggers collection of the metadata. While it does not
        # raise an error if metadata is not found while freezing, the calls below will fall at run-time in that case.
        dist = pkg_resources.get_distribution('my-test-package')

        # Sanity check
        assert dist.project_name == 'my-test-package'
        assert dist.version == '1.0'
        assert dist.egg_name() == f'my_test_package-{dist.version}-py{sys.version_info[0]}.{sys.version_info[1]}'
        """,
        pyi_args=['--paths', str(extra_path)]
    )


# Test that we can successfully package a program even if one of its modules contains non-ASCII characters in a local
# (non-UTF8) encoding and fails to declare such encoding using PEP361 encoding header.
#
# Python versions prior to 3.14.1 are able to import such modules; however, starting with python 3.14.1, a SyntaxError
# is raised. See: https://github.com/python/cpython/commit/9ff705c
@pytest.mark.skipif(sys.version_info >= (3, 14, 1), reason="python >= 3.14.1 disallows invalid characters")
def test_program_importing_module_with_invalid_encoding1(pyi_builder):
    # Add directory containing the my-test-package metadata to search path
    extra_path = _MODULES_DIR / "pyi_module_with_invalid_encoding"

    pyi_builder.test_source(
        """
        import mymodule1
        assert mymodule1.hello() == "hello"
        """,
        pyi_args=['--paths', str(extra_path)]
    )


@pytest.mark.skipif(sys.version_info >= (3, 14, 1), reason="python >= 3.14.1 disallows invalid characters")
def test_program_importing_module_with_invalid_encoding2(pyi_builder):
    # Add directory containing the my-test-package metadata to search path
    extra_path = _MODULES_DIR / "pyi_module_with_invalid_encoding"

    pyi_builder.test_source(
        """
        import mymodule2
        assert mymodule2.hello() == "hello"
        """,
        pyi_args=['--paths', str(extra_path)]
    )


# Test that collection of an executable shell script (essentially a data file with executable bit) preserves its
# executable bit.
@pytest.mark.linux
@pytest.mark.darwin
def test_bundled_shell_script(pyi_builder, tmp_path):
    script_file = tmp_path / "test_script.sh"
    with open(script_file, "w", encoding="utf-8") as fp:
        print('#!/bin/sh', file=fp)
        print('echo "Hello world!"', file=fp)
    script_file.chmod(0o755)

    pyi_builder.test_source(
        """
        import os
        import subprocess

        script = os.path.join(os.path.dirname(__file__), 'test_script.sh')
        output = subprocess.check_output(script, text=True)

        print(output)
        assert output.strip() == "Hello world!"
        """,
        pyi_args=['--add-data', f"{script_file}:."]
    )


# Test that a program importing `__main__` module does not pull in `PyInstaller` (or in the case of the test, the
# `pytest`). The problem is that the `__main__` has different meaning during analysis vs. during program's run;
# during analysis, it resolves to the entry-point module that is running the analysis, whereas during program run, it
# refers to the program's entry-point. Currently, this seems to be a problem only on Windows, where modulegraph manages
# to resolve `__main__` into `.../PyInstaller.exe/__main__.py` (or `.../pytest.exe/__main__.py`). On Linux and macOS,
# modulegraph does not seem to be able to resolve `__main__`.
def test_import_main_should_not_collect_pyinstaller1(pyi_builder):
    hooks_dir = _MODULES_DIR / 'pyi_import_main' / 'hooks'
    pyi_builder.test_source(
        """
        # Plain import.
        import __main__
        print(__main__)
        """,
        pyi_args=['--additional-hooks-dir', str(hooks_dir)]
    )


def test_import_main_should_not_collect_pyinstaller2(pyi_builder):
    hooks_dir = _MODULES_DIR / 'pyi_import_main' / 'hooks'
    pyi_builder.test_source(
        """
        # Import __main__ in the same way as `pkg_resources` and its vendored variants
        # (e.g., `pip._vendor.pkg_resources`) do.
        try:
            from __main__ import __requires__
        except ImportError:
            pass
        """,
        pyi_args=['--additional-hooks-dir', str(hooks_dir)]
    )


# Test that a relative import attempt of a missing optional sub-module in a package does not trigger collection of an
# unrelated but eponymous top-level module. Simulates the scenario from #8010, where the following block in
# `openpyxl.reader.excel`:
#
# ```
# try:
#    from ..tests import KEEP_VBA
# except ImportError:
#    KEEP_VBA = False
# ```
#
# (https://foss.heptapod.net/openpyxl/openpyxl/-/blob/branch/3.1/openpyxl/reader/excel.py#L16)
#
# triggers collection of top-level `tests` package that is provided by the `LaoNLP` distribution. And importing
# the said `tests` package during analysis triggers LaoNLP's unit tests...
def test_missing_relative_import_collects_unrelated_top_level_module(pyi_builder):
    extra_path = _MODULES_DIR / "pyi_missing_relative_import"
    hooks_dir = extra_path / 'hooks'

    pyi_builder.test_source(
        """
        import mypackage
        """,
        pyi_args=['--additional-hooks-dir', str(hooks_dir), '--paths', str(extra_path)]
    )  # yapf: disable


# Test that various forms of relative imports are properly caught by the module exclusion.
@pytest.mark.parametrize('exclude', [False, True], ids=["baseline", "exclude"])
def test_excluded_relative_imports(pyi_builder, exclude):
    extra_path = _MODULES_DIR / "pyi_excluded_relative_imports"
    hooks_dir = extra_path / 'hooks'

    pyi_args = ['--paths', str(extra_path)]
    if exclude:
        pyi_args += ['--additional-hooks-dir', str(hooks_dir)]

    pyi_builder.test_source(
        f"""
        import os
        os.environ['_FORBIDDEN_MODULES_ENABLED'] = '{str(int(not exclude))}'  # '0' or '1'

        import mypackage
        """,
        pyi_args=pyi_args,
    )


# Test the bytecode optimization settings (either implicit via python interpreter options or explicit via the new
# optimize option).
def _test_optimization(pyi_builder, level, tmp_path, pyi_args):
    extra_path = _MODULES_DIR / "pyi_optimization"
    results_filename = tmp_path / "results.json"

    pyi_args = ["--path", str(extra_path), *pyi_args]

    pyi_builder.test_script("pyi_optimization.py", pyi_args=pyi_args, app_args=[str(results_filename)])

    with open(results_filename, "r", encoding="utf-8") as fp:
        results = json.load(fp)

    # Check that sys.flags.optimize matches the specified level
    runtime_level = results["sys.flags.optimize"]
    assert runtime_level == level, \
        f"Unexpected sys.flags.optimize value! Expected {level}, found {runtime_level}!"

    def _check_flag(results, area, flag, expected_value):
        value = results[area][flag]
        assert value == expected_value, \
            f"Unexpected value for {flag!r} in {area!r}. Expected {expected_value!r}, found {value!r}"

    # Check results for entry-point script
    _check_flag(results, "script", "has_debug", level < 1)
    _check_flag(results, "script", "has_assert", level < 1)
    _check_flag(results, "script", "function_has_doc", level < 2)

    # Check results for module
    _check_flag(results, "module", "has_debug", level < 1)
    _check_flag(results, "module", "has_assert", level < 1)
    _check_flag(results, "module", "module_has_doc", level < 2)
    _check_flag(results, "module", "function_has_doc", level < 2)


@pytest.mark.parametrize('level', [0, 1, 2], ids=["unspecified", "O", "OO"])
def test_optimization_via_python_option(pyi_builder, level, tmp_path):
    pyi_args = level * ["--python", "O"]

    # If no "--python O" flags are supplied, the optimization level is set to `sys.flags.optimize`.
    if not pyi_args:
        level = sys.flags.optimize

    _test_optimization(pyi_builder, level, tmp_path, pyi_args)


@pytest.mark.parametrize('level', [0, 1, 2])
def test_optimization_via_optimize_option(pyi_builder, level, tmp_path):
    pyi_args = ["--optimize", str(level)]
    _test_optimization(pyi_builder, level, tmp_path, pyi_args)


# Test that runpy.run_path() in frozen application can run a bundled python script file. See #8767.
def test_runpy_run_from_location(tmp_path, pyi_builder):
    script_file = tmp_path / "script.py"
    script_file.write_text("""print("Hello world!")\n""", encoding='utf-8')

    pyi_builder.test_source(
        """
        import os
        import sys
        import runpy

        script_file = os.path.join(sys._MEIPASS, 'script.py')
        runpy.run_path(script_file, run_name="__main__")
        """,
        pyi_args=['--add-data', f"{script_file}:."],
    )


# End-to-end tests with --add-data and source paths with and without glob.
@pytest.mark.parametrize('scenario', ['recursive_noglob', 'recursive_glob', 'top_level_files'])
def test_recursive_add_data(pyi_builder, scenario):
    data_dir = pathlib.Path(__file__).parent / 'data' / 'recursive_add_data' / 'data-dir'

    if scenario == 'top_level_files':
        add_data_arg = data_dir / '*.txt'
        expected_files = [
            "data-dir/file1.txt",
            "data-dir/file2.txt",
        ]
    else:
        if scenario == 'recursive_noglob':
            add_data_arg = data_dir
        else:
            add_data_arg = data_dir / '*'
        expected_files = [
            "data-dir/dir1/dir1/file1.txt",
            "data-dir/dir1/dir1/file2.txt",
            "data-dir/dir1/dir2/file1.txt",
            "data-dir/dir1/file1.txt",
            "data-dir/dir2/dir1/file1.txt",
            "data-dir/dir2/dir1/file2.txt",
            "data-dir/dir3/dir1/file1.txt",
            "data-dir/dir3/file1.txt",
            "data-dir/dir3/file2.txt",
            "data-dir/file1.txt",
            "data-dir/file2.txt",
        ]

    pyi_builder.test_source(
        """
        import pathlib
        import sys

        root_directory = pathlib.Path(__file__).parent

        # File names to check are passed as command-line arguments; each
        # file is expected to contain its relative path (with POSIX separators).
        for entry in sys.argv[1:]:
            file_path = root_directory / entry
            assert file_path.is_file(), f"File {str(file_path)!r} does not exist!"

            content = file_path.read_text().strip()
            assert content == entry, f"Unexpected content in {str(file_path)!r}: found {content!r}, expected {entry!r}!"
        """,
        pyi_args=['--add-data', f'{add_data_arg!s}:data-dir'],
        app_args=[*expected_files],
    )


# Ensure that time.sleep() works as expected in the frozen application. See #8104, #9225.
# Separately test console/windowed builds, as they have (slightly) different bootloaders.
@pytest.mark.parametrize('windowed', [False, True], ids=['console', 'windowed'])
def test_time_sleep(pyi_builder, windowed):
    pyi_builder.test_source(
        """
        import sys
        import time
        import datetime

        ITERATIONS = 5
        DELAY = 1  # seconds
        TOL = 50  # milliseconds

        elapsed = []
        for i in range(ITERATIONS):
            print(f"Iteration #{i+1} at {datetime.datetime.now()}", file=sys.stderr)
            start_time = time.monotonic()
            time.sleep(DELAY)
            elapsed.append(time.monotonic() - start_time)

        print("Elapsed times (monotonic clock):", file=sys.stderr)

        test_ok = True
        for idx, value in enumerate(elapsed):
            delta = (value - DELAY) * 1000  # ms

            # We are trying to catch cases when the elapsed time interval is *shorter* than the requested delay, which
            # indicates mis-behaving time.sleep() as per #8104. Typically the elapsed time interval is a bit longer than
            # the requested delay, but the delta varies depending on system scheduling and load.
            if delta < 0 and abs(delta) >= TOL:
                status = 'TOO SHORT'
                test_ok = False
            else:
                status = 'OK'

            print(f" - #{idx+1}: {value:.6f} s, delta: {delta:.4f} ms ({status})", file=sys.stderr)

        if test_ok:
            print("Test passed.", file=sys.stderr)
        else:
            print("Test failed.", file=sys.stderr)
            sys.exit(1)
        """,
        pyi_args=['--windowed'] if windowed else [],
    )


# Test that on Windows, values of dwFlags and wShowWindow fields in STATRTUPINFO structure are propagated into onefile
# child process. Onedir variant serves as sanity check. See #9342.
@pytest.mark.win32
@pytest.mark.parametrize('windowed', [False, True], ids=['console', 'windowed'])
def test_startupinfo_flags(pyi_builder, tmp_path, windowed):
    pyi_builder.test_script(
        'pyi_get_startupinfo_flags.py',
        pyi_args=['--windowed'] if windowed else [],
    )

    # Find executable
    exes = pyi_builder._find_executables('pyi_get_startupinfo_flags')
    assert len(exes) == 1
    program_exe = exes[0]

    # Re-run with various flag combinations
    TEST_FLAGS = (
        (False, 0),
        (True, 0),  # SW_HIDE
        (True, 1),  # SW_NORMAL
        (True, 6),  # SW_MINIMIZE
    )

    for i in range(len(TEST_FLAGS)):
        enabled, value = TEST_FLAGS[i]
        print(
            f"=== running test variant #{i + 1}: STARTF_USESHOWWINDOW={enabled}, wShowWindow={value} ===",
            file=sys.stderr,
        )

        output_file = tmp_path / f'output{i}.json'

        si = subprocess.STARTUPINFO()
        if enabled:
            si.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            si.wShowWindow = value

        subprocess.run([program_exe, output_file], check=True, startupinfo=si)

        with open(output_file, 'r') as fp:
            results = json.load(fp)

        if enabled:
            assert (results['dwFlags'] & subprocess.STARTF_USESHOWWINDOW) != 0, \
                'dwFlags does not contain STARTF_USESHOWWINDOW, but it should!'
            assert results['wShowWindow'] == value, \
                f"Unexpected wShowWindow value - expected {value}, found {results['wShowWindow']}!"
        else:
            assert (results['dwFlags'] & subprocess.STARTF_USESHOWWINDOW) == 0, \
                'dwFlags contains STARTF_USESHOWWINDOW, but it should not!'



import os
import sys


def setup_vendored_packages(mode):
    # Resolve _vendored directory within this package.
    vendored_path = os.path.join(os.path.dirname(__file__), '_vendored')

    # Prepend or append the vendored path to sys.path.
    if mode == 'prepend':
        sys.path.insert(0, vendored_path)
    elif mode == 'append':
        sys.path.append(vendored_path)
    else:
        raise ValueError(f"Invalid mode: {mode}")



# -----------------------------------------------------------------------------
# Copyright (c) 2024, PyInstaller Development Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------

import os
import importlib
import atexit

# Helper for ensuring that only one Qt bindings package is registered at run-time via run-time hooks.
_registered_qt_bindings = None


def ensure_single_qt_bindings_package(qt_bindings):
    global _registered_qt_bindings
    if _registered_qt_bindings is not None:
        raise RuntimeError(
            f"Cannot execute run-time hook for {qt_bindings!r} because run-time hook for {_registered_qt_bindings!r} "
            "has been run before, and PyInstaller-frozen applications do not support multiple Qt bindings in the same "
            "application!"
        )
    _registered_qt_bindings = qt_bindings


# Helper for relocating Qt prefix via embedded qt.conf file.
_QT_CONF_FILENAME = ":/qt/etc/qt.conf"

_QT_CONF_RESOURCE_NAME = (
    # qt
    b"\x00\x02"
    b"\x00\x00\x07\x84"
    b"\x00\x71"
    b"\x00\x74"
    # etc
    b"\x00\x03"
    b"\x00\x00\x6c\xa3"
    b"\x00\x65"
    b"\x00\x74\x00\x63"
    # qt.conf
    b"\x00\x07"
    b"\x08\x74\xa6\xa6"
    b"\x00\x71"
    b"\x00\x74\x00\x2e\x00\x63\x00\x6f\x00\x6e\x00\x66"
)

_QT_CONF_RESOURCE_STRUCT = (
    # :
    b"\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x01"
    # :/qt
    b"\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x02"
    # :/qt/etc
    b"\x00\x00\x00\x0a\x00\x02\x00\x00\x00\x01\x00\x00\x00\x03"
    # :/qt/etc/qt.conf
    b"\x00\x00\x00\x16\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00"
)


def create_embedded_qt_conf(qt_bindings, prefix_path):
    # The QtCore module might be unavailable if we collected just the top-level binding package (e.g., PyQt5) without
    # any of its submodules. Since this helper is called from run-time hook for the binding package, we need to handle
    # that scenario here.
    try:
        QtCore = importlib.import_module(qt_bindings + ".QtCore")
    except ImportError:
        return

    # No-op if embedded qt.conf already exists
    if QtCore.QFile.exists(_QT_CONF_FILENAME):
        return

    # Create qt.conf file that relocates Qt prefix.
    # NOTE: paths should use POSIX-style forward slashes as separator, even on Windows.
    if os.sep == '\\':
        prefix_path = prefix_path.replace(os.sep, '/')

    qt_conf = f"[Paths]\nPrefix = {prefix_path}\n"
    if os.name == 'nt' and qt_bindings in {"PySide2", "PySide6"}:
        # PySide PyPI wheels on Windows set LibraryExecutablesPath to PrefixPath
        qt_conf += f"LibraryExecutables = {prefix_path}"

    # Encode the contents; in Qt5, QSettings uses Latin1 encoding, in Qt6, it uses UTF8.
    if qt_bindings in {"PySide2", "PyQt5"}:
        qt_conf = qt_conf.encode("latin1")
    else:
        qt_conf = qt_conf.encode("utf-8")

    # Prepend data size (32-bit integer, big endian)
    qt_conf_size = len(qt_conf)
    qt_resource_data = qt_conf_size.to_bytes(4, 'big') + qt_conf

    # Register
    succeeded = QtCore.qRegisterResourceData(
        0x01,
        _QT_CONF_RESOURCE_STRUCT,
        _QT_CONF_RESOURCE_NAME,
        qt_resource_data,
    )
    if not succeeded:
        return  # Tough luck

    # Unregister the resource at exit, to ensure that the registered resource on Qt/C++ side does not outlive the
    # `_qt_resource_data` python variable and its data buffer. This also adds a reference to the `_qt_resource_data`,
    # which conveniently ensures that the data is not garbage collected before we perform the cleanup (otherwise garbage
    # collector might kick in at any time after we exit this helper function, and `qRegisterResourceData` does not seem
    # to make a copy of the data!).
    atexit.register(
        QtCore.qUnregisterResourceData,
        0x01,
        _QT_CONF_RESOURCE_STRUCT,
        _QT_CONF_RESOURCE_NAME,
        qt_resource_data,
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2021-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------


def _test_basic_imports():
    # Import a very simple and rarely used pure-python lib ...
    import getopt  # noqa: F401
    # ... and a module importing a shared lib.
    import ssl  # noqa: F401

    print('Hello World!')


def _test_nested_data_file():
    # Try reading secret from a file in sub-directory.
    import os

    secret_file = os.path.join(__path__[0], 'data', 'secret.txt')
    print("Reading secret from %s..." % secret_file)
    with open(secret_file, 'r') as fp:
        secret = fp.read().strip()
    print("Secret: %s" % secret)

    assert secret == 'Secret1234'


def _test_nested_extensions():
    # Import psutil, which contains an extension in its package directory.
    import psutil  # noqa: F401
    print("Successfully imported psutil!")


def test_function():
    # Main test function
    _test_basic_imports()
    _test_nested_data_file()
    _test_nested_extensions()



#-----------------------------------------------------------------------------
# Copyright (c) 2017-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller.utils.hooks import collect_submodules, check_requirement

# Pandas keeps Python extensions loaded with dynamic imports here.
hiddenimports = collect_submodules('pandas._libs')

# Pandas 1.2.0 and later require cmath hidden import on linux and macOS. On Windows, this is not strictly required, but
# we add it anyway to keep things simple (and future-proof).
if check_requirement('pandas >= 1.2.0'):
    hiddenimports += ['cmath']



from setuptools import setup

setup(
    name="nspkg",
    version="1.0",
    namespace_packages=['nspkg', 'nspkg.nssubpkg'],
    packages=['nspkg', 'nspkg.nssubpkg'],
    package_dir = {'': 'src'},
    zip_safe=False,
)



#-----------------------------------------------------------------------------
# Copyright (c) 2025, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller.utils.hooks import is_module_satisfies

# As of SciPy 1.16.0, `scipy.spatial._ckdtree` extension started to depend on newly-introduced `scipy._cyutility`.
if is_module_satisfies('scipy >= 1.16.0'):
    hiddenimports = ['scipy._cyutility']



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import os, shlex, sys
from waflib.TaskGen import feature, after_method, taskgen_method
from waflib import Utils, Task, Logs, Options
from waflib.Tools import ccroot

testlock = Utils.threading.Lock()
SCRIPT_TEMPLATE = """#! %(python)s
import subprocess, sys
cmd = %(cmd)r
# if you want to debug with gdb:
#cmd = ['gdb', '-args'] + cmd
env = %(env)r
status = subprocess.call(cmd, env=env, cwd=%(cwd)r, shell=isinstance(cmd, str))
sys.exit(status)
"""


@taskgen_method
def handle_ut_cwd(self, key):
    cwd = getattr(self, key, None)
    if cwd:
        if isinstance(cwd, str):
            if os.path.isabs(cwd):
                self.ut_cwd = self.bld.root.make_node(cwd)
            else:
                self.ut_cwd = self.path.make_node(cwd)


@feature('test_scripts')
def make_interpreted_test(self):
    for x in ['test_scripts_source', 'test_scripts_template']:
        if not hasattr(self, x):
            Logs.warn('a test_scripts taskgen i missing %s' % x)
            return
    self.ut_run, lst = Task.compile_fun(self.test_scripts_template, shell=getattr(self, 'test_scripts_shell', False))
    script_nodes = self.to_nodes(self.test_scripts_source)
    for script_node in script_nodes:
        tsk = self.create_task('utest', [script_node])
        tsk.vars = lst + tsk.vars
        tsk.env['SCRIPT'] = script_node.path_from(tsk.get_cwd())
    self.handle_ut_cwd('test_scripts_cwd')
    env = getattr(self, 'test_scripts_env', None)
    if env:
        self.ut_env = env
    else:
        self.ut_env = dict(os.environ)
    paths = getattr(self, 'test_scripts_paths', {})
    for (k, v) in paths.items():
        p = self.ut_env.get(k, '').split(os.pathsep)
        if isinstance(v, str):
            v = v.split(os.pathsep)
        self.ut_env[k] = os.pathsep.join(p + v)


@feature('test')
@after_method('apply_link', 'process_use')
def make_test(self):
    if not getattr(self, 'link_task', None):
        return
    tsk = self.create_task('utest', self.link_task.outputs)
    if getattr(self, 'ut_str', None):
        self.ut_run, lst = Task.compile_fun(self.ut_str, shell=getattr(self, 'ut_shell', False))
        tsk.vars = lst + tsk.vars
    self.handle_ut_cwd('ut_cwd')
    if not hasattr(self, 'ut_paths'):
        paths = []
        for x in self.tmp_use_sorted:
            try:
                y = self.bld.get_tgen_by_name(x).link_task
            except AttributeError:
                pass
            else:
                if not isinstance(y, ccroot.stlink_task):
                    paths.append(y.outputs[0].parent.abspath())
        self.ut_paths = os.pathsep.join(paths) + os.pathsep
    if not hasattr(self, 'ut_env'):
        self.ut_env = dct = dict(os.environ)

        def add_path(var):
            dct[var] = self.ut_paths + dct.get(var, '')

        if Utils.is_win32:
            add_path('PATH')
        elif Utils.unversioned_sys_platform() == 'darwin':
            add_path('DYLD_LIBRARY_PATH')
            add_path('LD_LIBRARY_PATH')
        else:
            add_path('LD_LIBRARY_PATH')
    if not hasattr(self, 'ut_cmd'):
        self.ut_cmd = getattr(Options.options, 'testcmd', False)


@taskgen_method
def add_test_results(self, tup):
    Logs.debug("ut: %r", tup)
    try:
        self.utest_results.append(tup)
    except AttributeError:
        self.utest_results = [tup]
    try:
        self.bld.utest_results.append(tup)
    except AttributeError:
        self.bld.utest_results = [tup]


@Task.deep_inputs
class utest(Task.Task):
    color = 'PINK'
    after = ['vnum', 'inst']
    vars = []

    def runnable_status(self):
        if getattr(Options.options, 'no_tests', False):
            return Task.SKIP_ME
        ret = super(utest, self).runnable_status()
        if ret == Task.SKIP_ME:
            if getattr(Options.options, 'all_tests', False):
                return Task.RUN_ME
        return ret

    def get_test_env(self):
        return self.generator.ut_env

    def post_run(self):
        super(utest, self).post_run()
        if getattr(Options.options, 'clear_failed_tests', False) and self.waf_unit_test_results[1]:
            self.generator.bld.task_sigs[self.uid()] = None

    def run(self):
        if hasattr(self.generator, 'ut_run'):
            return self.generator.ut_run(self)
        self.ut_exec = getattr(self.generator, 'ut_exec', [self.inputs[0].abspath()])
        ut_cmd = getattr(self.generator, 'ut_cmd', False)
        if ut_cmd:
            self.ut_exec = shlex.split(ut_cmd % ' '.join(self.ut_exec))
        return self.exec_command(self.ut_exec)

    def exec_command(self, cmd, **kw):
        self.generator.bld.log_command(cmd, kw)
        if getattr(Options.options, 'dump_test_scripts', False):
            script_code = SCRIPT_TEMPLATE % {
                'python': sys.executable,
                'env': self.get_test_env(),
                'cwd': self.get_cwd().abspath(),
                'cmd': cmd
            }
            script_file = self.inputs[0].abspath() + '_run.py'
            Utils.writef(script_file, script_code, encoding='utf-8')
            os.chmod(script_file, Utils.O755)
            if Logs.verbose > 1:
                Logs.info('Test debug file written as %r' % script_file)
        proc = Utils.subprocess.Popen(
            cmd,
            cwd=self.get_cwd().abspath(),
            env=self.get_test_env(),
            stderr=Utils.subprocess.PIPE,
            stdout=Utils.subprocess.PIPE,
            shell=isinstance(cmd, str)
        )
        (stdout, stderr) = proc.communicate()
        self.waf_unit_test_results = tup = (self.inputs[0].abspath(), proc.returncode, stdout, stderr)
        testlock.acquire()
        try:
            return self.generator.add_test_results(tup)
        finally:
            testlock.release()

    def get_cwd(self):
        return getattr(self.generator, 'ut_cwd', self.inputs[0].parent)


def summary(bld):
    lst = getattr(bld, 'utest_results', [])
    if lst:
        Logs.pprint('CYAN', 'execution summary')
        total = len(lst)
        tfail = len([x for x in lst if x[1]])
        Logs.pprint('GREEN', '  tests that pass %d/%d' % (total - tfail, total))
        for (f, code, out, err) in lst:
            if not code:
                Logs.pprint('GREEN', '    %s' % f)
        Logs.pprint('GREEN' if tfail == 0 else 'RED', '  tests that fail %d/%d' % (tfail, total))
        for (f, code, out, err) in lst:
            if code:
                Logs.pprint('RED', '    %s' % f)


def set_exit_code(bld):
    lst = getattr(bld, 'utest_results', [])
    for (f, code, out, err) in lst:
        if code:
            msg = []
            if out:
                msg.append('stdout:%s%s' % (os.linesep, out.decode('utf-8')))
            if err:
                msg.append('stderr:%s%s' % (os.linesep, err.decode('utf-8')))
            bld.fatal(os.linesep.join(msg))


def options(opt):
    opt.add_option('--notests', action='store_true', default=False, help='Exec no unit tests', dest='no_tests')
    opt.add_option('--alltests', action='store_true', default=False, help='Exec all unit tests', dest='all_tests')
    opt.add_option(
        '--clear-failed',
        action='store_true',
        default=False,
        help='Force failed unit tests to run again next time',
        dest='clear_failed_tests'
    )
    opt.add_option(
        '--testcmd',
        action='store',
        default=False,
        dest='testcmd',
        help=
        'Run the unit tests using the test-cmd string example "--testcmd="valgrind --error-exitcode=1 %s" to run under valgrind'
    )
    opt.add_option(
        '--dump-test-scripts',
        action='store_true',
        default=False,
        help='Create python scripts to help debug tests',
        dest='dump_test_scripts'
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2021-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
#
# Tests for stdlib `inspect` module.

import pathlib

import pytest

# Directory with testing modules used in some tests.
_MODULES_DIR = pathlib.Path(__file__).parent / 'modules'


def _patch_collection_mode(monkeypatch, module_name):
    # Patch Analysis to set module_collection_mode for specified package/module.
    import PyInstaller.building.build_main

    class _Analysis(PyInstaller.building.build_main.Analysis):
        def __init__(self, *args, **kwargs):
            kwargs['module_collection_mode'] = {
                module_name: 'pyz+py',
            }
            super().__init__(*args, **kwargs)

    monkeypatch.setattr('PyInstaller.building.build_main.Analysis', _Analysis)


# Test that we can retrieve source for a module that was collected both into PYZ archive and as a source .py file.
@pytest.mark.parametrize('module_name', ['mypackage', 'mypackage.mod_a'], ids=['package', 'submodule'])
def test_inspect_getsource(pyi_builder, module_name, monkeypatch):
    pathex = _MODULES_DIR / 'pyi_inspect_getsource'
    _patch_collection_mode(monkeypatch, 'mypackage')
    pyi_builder.test_source(
        f"""
        import inspect

        import {module_name}

        # Retrieve source via module instance.
        source = inspect.getsource({module_name})
        print(source)

        # Check that the source starts with expected comment
        EXPECTED_COMMENT = "# {module_name}: "
        if not source.startswith(EXPECTED_COMMENT):
            raise ValueError(f"Source does not start with expected comment: {{EXPECTED_COMMENT}}")
        """,
        pyi_args=['--paths', str(pathex)],
    )


def test_inspect_getsource_class_from_base_library_module(pyi_builder, monkeypatch):
    _patch_collection_mode(monkeypatch, 'enum')
    pyi_builder.test_source(
        """
        import sys
        import os
        import enum
        import inspect

        # Ensure that module is collected in `base_library.zip`; normalize separators before comparison.
        BASE_LIBRARY_ZIP = os.path.normpath(os.path.join(sys._MEIPASS, 'base_library.zip'))
        if not enum.__file__.startswith(BASE_LIBRARY_ZIP):
            raise ValueError(f"The 'enum' module is not collected in base_library.zip: {enum.__file__}")

        # Retrieve source via class (enum.Enum)
        source = inspect.getsource(enum.Enum)
        print(source)
        """
    )


def test_inspect_getsource_class_method_from_base_library_module(pyi_builder, monkeypatch):
    _patch_collection_mode(monkeypatch, 'enum')
    pyi_builder.test_source(
        """
        import sys
        import os
        import enum
        import inspect

        # Ensure that module is collected in `base_library.zip`; normalize separators before comparison.
        BASE_LIBRARY_ZIP = os.path.normpath(os.path.join(sys._MEIPASS, 'base_library.zip'))
        if not enum.__file__.startswith(BASE_LIBRARY_ZIP):
            raise ValueError(f"The 'enum' module is not collected in base_library.zip: {enum.__file__}")

        # Retrieve source via class method (enum.Enum.__new__)
        source = inspect.getsource(enum.Enum.__new__)
        print(source)
        """
    )


# Similar to `test_inspect_getsource`, except that we are retrieving source for a function within the module.
def test_inspect_getsource_function(pyi_builder, monkeypatch):
    pathex = _MODULES_DIR / 'pyi_inspect_getsource'
    _patch_collection_mode(monkeypatch, 'mypackage')
    pyi_builder.test_source(
        """
        import inspect

        from mypackage.mod_b import test_function

        # Retrieve source for function.
        source = inspect.getsource(test_function)
        print(source)

        # Check that the source starts with function definition
        EXPECTED_START = "def test_function():"
        if not source.startswith(EXPECTED_START):
            raise ValueError(f"Source does not start with function definition: {EXPECTED_START}")

        # Check that the comment is preset.
        EXPECTED_COMMENT = "# A comment."
        if EXPECTED_COMMENT not in source:
            raise ValueError(f"Source does not contain expected comment: {EXPECTED_COMMENT}")
        """,
        pyi_args=['--paths', str(pathex)],
    )


# Test inspect.getmodule() on stack-frames obtained by inspect.stack(). Reproduces the issue reported by #5963 while
# expanding the test to cover a package and its submodule in addition to the __main__ module.
def test_inspect_getmodule_from_stackframes(pyi_builder):
    pathex = _MODULES_DIR / 'pyi_inspect_getmodule_from_stackframes'
    # NOTE: run_from_path MUST be True, otherwise cwd + rel_path coincides with sys._MEIPASS + rel_path and masks the
    # path resolving issue in onedir builds.
    pyi_builder.test_source(
        """
        import helper_package

        # helper_package.test_call_chain() calls eponymous function in helper_package.helper_module, which in turn uses
        # inspect.stack() and inspect.getmodule() to obtain list of modules involved in the chain call.
        modules = helper_package.test_call_chain()

        # Expected call chain
        expected_module_names = [
            'helper_package.helper_module',
            'helper_package',
            '__main__'
        ]

        # All modules must have been resolved
        assert not any(module is None for module in modules)

        # Verify module names
        module_names = [module.__name__ for module in modules]
        assert module_names == expected_module_names
        """,
        pyi_args=['--paths', str(pathex)],
        run_from_path=True
    )


# Test the robustness of `inspect` run-time hook w.r.t. to the issue #7642.
#
# If our run-time hook imports a module in the global namespace and attempts to use this module in a function that
# might get called later on in the program (e.g., a function override or registered callback function), we are at the
# mercy of user's program, which might re-bind the module's name to something else (variable, function), leading to
# an error.
#
# This particular test will raise:
# ```
# Traceback (most recent call last):
#  File "test_source.py", line 17, in <module>
#  File "test_source.py", line 14, in some_interactive_debugger_function
#  File "inspect.py", line 1755, in stack
#  File "inspect.py", line 1730, in getouterframes
#  File "inspect.py", line 1688, in getframeinfo
#  File "PyInstaller/hooks/rthooks/pyi_rth_inspect.py", line 22, in _pyi_getsourcefile
# AttributeError: 'function' object has no attribute 'getfile'
# ```
def test_inspect_rthook_robustness(pyi_builder):
    pyi_builder.test_source(
        """
        # A custom function in global namespace that happens to have name clash with `inspect` module.
        def inspect(something):
            print(f"Inspecting {something}: type is {type(something)}")


        # A call to `inspect.stack` function somewhere deep in an interactive debugger framework.
        # This eventually ends up calling our `_pyi_getsourcefile` override in the `inspect` run-time hook. The
        # override calls `inspect.getfile`; if the run-time hook imported `inspect` in a global namespace, the
        # name at this point is bound the the custom function that program defined, leading to an error.
        def some_interactive_debugger_function():
            import inspect
            print(f"Current stack: {inspect.stack()}")


        some_interactive_debugger_function()
        """
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2005-2023, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------
"""
Functional tests for Matplotlib.
"""

import pytest

from PyInstaller.utils.tests import importorskip
from PyInstaller.utils.hooks import check_requirement

# List of tuples "(backend_name, qt_bindings)", where:
#
# * "backend_name" is the name of a Matplotlib backend to be tested below.
# * "qt_bindings" is the name of the external Qt bindings package required by this backend.
#
if check_requirement("matplotlib >= 3.5.0"):
    # Matplotlib 3.5.0 introduced a unified Qt backend that supports PySide2, PyQt5, PySide6, and PyQt6.
    _backends = [
        ('QtAgg', 'PyQt5'),
        ('QtAgg', 'PySide2'),
        ('QtAgg', 'PyQt6'),
        ('QtAgg', 'PySide6'),
    ]
else:
    _backends = [
        ('Qt5Agg', 'PyQt5'),
        ('Qt5Agg', 'PySide2'),
    ]


# Test Matplotlib with access to only one backend at a time.
@importorskip('matplotlib')
@pytest.mark.parametrize(
    'backend_name, qt_bindings',
    [
        pytest.param(backend_name, qt_bindings, marks=importorskip(qt_bindings))
        for backend_name, qt_bindings in _backends
    ],
    ids=[qt_bindings for backend_name, qt_bindings in _backends],
)
def test_matplotlib(pyi_builder, monkeypatch, backend_name, qt_bindings):
    '''
    Test Matplotlib with the passed backend enabled, the passed backend package included with this frozen application,
    all other backend packages explicitly excluded from this frozen application, and the passed rcParam key set to the
    corresponding passed value if that key is _not_ `None` or ignore that value otherwise.
    '''

    # Exclude all Qt bindings except the ones we are using in this test.
    pyi_args = [
        f'--exclude-module={bindings_name}' for backend_name, bindings_name in _backends if bindings_name != qt_bindings
    ]

    # Test program
    pyi_builder.test_source(
        f"""
        import os
        import sys
        import tempfile

        import matplotlib

        # Matplotlib >= v3.4.0 allows Qt bindings name in QT_API environment variable to be capitalized. Lower-case it
        # here just in case we ever happen to run the test with older version.
        qt_bindings_lower = {qt_bindings!r}.lower()

        # Report these parameters.
        print(f'Testing Matplotlib with backend={backend_name} and QT_API={{qt_bindings_lower}}')

        # Configure Matplotlib *BEFORE* calling any Matplotlib functions.
        matplotlib.rcParams['backend'] = {backend_name!r}
        os.environ['QT_API'] = qt_bindings_lower

        # Enable the desired backend *BEFORE* plotting with this backend.
        matplotlib.use({backend_name!r})

        # A runtime hook should force Matplotlib to create its configuration directory in a temporary directory
        # rather than in $HOME/.matplotlib.
        configdir = os.environ['MPLCONFIGDIR']
        print(f'MPLCONFIGDIR: {{configdir}}')
        if not configdir.startswith(tempfile.gettempdir()):
            raise SystemExit('MPLCONFIGDIR not pointing to temp directory.')

        # Test access to the standard 'mpl_toolkits' namespace package installed with Matplotlib.
        # Note that this import was reported to fail under Matplotlib 1.3.0.
        from mpl_toolkits import axes_grid1

        # Try importing pyplot. This will attempt to activate the selected backend.
        from matplotlib import pyplot as plt
        """,
        pyi_args=pyi_args,
    )



#-----------------------------------------------------------------------------
# Copyright (c) 2025, PyInstaller Development Team.
#
# Distributed under the terms of the GNU General Public License (version 2
# or later) with exception for distributing the bootloader.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: (GPL-2.0-or-later WITH Bootloader-exception)
#-----------------------------------------------------------------------------

from PyInstaller import compat
from PyInstaller.utils import hooks as hookutils


def pre_safe_import_module(api):
    if compat.is_linux:
        # See comment in the adjacent `hook-gi.py`.
        try:
            paths = hookutils.get_module_attribute(api.module_name, "__path__")
        except Exception:
            # Most likely `gi.overrides` cannot be imported.
            paths = []

        for path in paths:
            api.append_package_path(path)



#! /usr/bin/env python
# encoding: utf-8
# WARNING! Do not edit! https://waf.io/book/index.html#_obtaining_the_waf_file

import re

INC_REGEX = r"""(?:^|['">]\s*;)\s*(?:|#\s*)INCLUDE\s+(?:\w+_)?[<"'](.+?)(?=["'>])"""
USE_REGEX = r"""(?:^|;)\s*USE(?:\s+|(?:(?:\s*,\s*(?:NON_)?INTRINSIC)?\s*::))\s*(\w+)"""
MOD_REGEX = r"""(?:^|;)\s*MODULE(?!\s+(?:PROCEDURE|SUBROUTINE|FUNCTION))\s+(\w+)"""
SMD_REGEX = r"""(?:^|;)\s*SUBMODULE\s*\(([\w:]+)\)\s*(\w+)"""
re_inc = re.compile(INC_REGEX, re.I)
re_use = re.compile(USE_REGEX, re.I)
re_mod = re.compile(MOD_REGEX, re.I)
re_smd = re.compile(SMD_REGEX, re.I)


class fortran_parser(object):
    def __init__(self, incpaths):
        self.seen = []
        self.nodes = []
        self.names = []
        self.incpaths = incpaths

    def find_deps(self, node):
        txt = node.read()
        incs = []
        uses = []
        mods = []
        for line in txt.splitlines():
            m = re_inc.search(line)
            if m:
                incs.append(m.group(1))
            m = re_use.search(line)
            if m:
                uses.append(m.group(1))
            m = re_mod.search(line)
            if m:
                mods.append(m.group(1))
            m = re_smd.search(line)
            if m:
                uses.append(m.group(1))
                mods.append('{0}:{1}'.format(m.group(1), m.group(2)))
        return (incs, uses, mods)

    def start(self, node):
        self.waiting = [node]
        while self.waiting:
            nd = self.waiting.pop(0)
            self.iter(nd)

    def iter(self, node):
        incs, uses, mods = self.find_deps(node)
        for x in incs:
            if x in self.seen:
                continue
            self.seen.append(x)
            self.tryfind_header(x)
        for x in uses:
            name = "USE@%s" % x
            if not name in self.names:
                self.names.append(name)
        for x in mods:
            name = "MOD@%s" % x
            if not name in self.names:
                self.names.append(name)

    def tryfind_header(self, filename):
        found = None
        for n in self.incpaths:
            found = n.find_resource(filename)
            if found:
                self.nodes.append(found)
                self.waiting.append(found)
                break
        if not found:
            if not filename in self.names:
                self.names.append(filename)
